from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

import numpy as np  # noqa: TCH002

from ropt import utils
from ropt.enums import ResultAxisName

from ._result_field import ResultField
from ._utils import (
    _get_nonlinear_constraint_values,
    _get_nonlinear_constraint_violations,
    _immutable_copy,
)

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig

    from ._functions import Functions


@dataclass
class NonlinearConstraints(ResultField):
    """This class stores constraint values and violations.

    The following information is stored:

    1. Differences between the non-linear constraints and their right-hand-side
       values.
    2. Violations of constraints, defined as the absolute value of the
       difference if the constraint is violated, or else zero.
    3. Scaled versions for these values.

    Args:
        values:            Non-linear constraint values
        violations:        Violations of the nonlinear constraints
        scaled_values:     Optional scaled non-linear constraint values
        scaled_violations: Optional scaled violations of the nonlinear constraints
    """

    values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={"__axes__": (ResultAxisName.NONLINEAR_CONSTRAINT,)},
    )
    violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.NONLINEAR_CONSTRAINT,),
        },
    )
    scaled_values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.NONLINEAR_CONSTRAINT,),
        },
    )
    scaled_violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.NONLINEAR_CONSTRAINT,),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.values = _immutable_copy(self.values)
        self.violations = _immutable_copy(self.violations)
        self.scaled_values = _immutable_copy(self.scaled_values)
        self.scaled_violations = _immutable_copy(self.scaled_violations)

    @classmethod
    def create(
        cls,
        config: EnOptConfig,
        functions: Optional[Functions],
        constraint_auto_scales: Optional[Optional[NDArray[np.float64]]],
    ) -> Optional[NonlinearConstraints]:
        """Add constraint information.

        This factory function creates a `NonlinearConstraints` object with the
        correct information on nonlinear constraints, based on the given
        evaluations and functions.

        Args:
            config:                 The ensemble optimizer configuration object
            functions:              An instance of the Function class
            constraint_auto_scales: Optional constraint scales
        Returns:
            A newly created NonlinearConstraints object.
        """
        if functions is None:
            return None

        def _get_scaled(
            unscaled: Optional[NDArray[np.float64]],
        ) -> Optional[NDArray[np.float64]]:
            if unscaled is not None:
                assert config is not None
                return utils.scaling.scale_constraints(
                    config,
                    unscaled,
                    None if constraint_auto_scales is None else constraint_auto_scales,
                    axis=-1,
                )
            return None

        if config.nonlinear_constraints is not None:
            assert functions.constraints is not None
            values = _get_nonlinear_constraint_values(
                config, functions.constraints, axis=-1
            )
            violations = _get_nonlinear_constraint_violations(config, values)
            scaled_values = _get_scaled(values)
            scaled_violations = _get_scaled(violations)
            return NonlinearConstraints(
                values=values,
                violations=violations,
                scaled_values=scaled_values,
                scaled_violations=scaled_violations,
            )

        return None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

import numpy as np

from ropt import utils
from ropt.enums import ResultAxisName

from ._result_field import ResultField
from ._utils import (
    _get_lower_bound_constraint_values,
    _get_lower_bound_violations,
    _get_upper_bound_constraint_values,
    _get_upper_bound_violations,
    _immutable_copy,
)

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig

    from ._function_evaluations import FunctionEvaluations


@dataclass
class BoundConstraints(ResultField):
    """This class stores bound constraint values and violations.

    The following information is stored:

    1. Differences between the current variable and the lower or upper bounds.
    2. Violations of constraints, defined as the absolute value of the
       difference if the constraint is violated, or else zero.
    3. If variable scaling is configured, unscaled versions of all values.

    Args:
        lower_values:              Lower bound differences
        lower_violations:          Lower bound violations
        upper_values:              Upper bound differences
        upper_violations:          Upper bound violations
        unscaled_lower_values:     Optional unscaled lower bound differences
        unscaled_lower_violations: Optional unscaled lower bound violations
        unscaled_upper_values:     Optional unscaled upper bound differences
        unscaled_upper_violations: Optional unscaled upper bound violations
    """

    lower_values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    lower_violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    upper_values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    upper_violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    unscaled_lower_values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    unscaled_lower_violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    unscaled_upper_values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    unscaled_upper_violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.lower_values = _immutable_copy(self.lower_values)
        self.lower_violations = _immutable_copy(self.lower_violations)
        self.upper_values = _immutable_copy(self.upper_values)
        self.upper_violations = _immutable_copy(self.upper_violations)
        self.unscaled_lower_values = _immutable_copy(self.unscaled_lower_values)
        self.unscaled_lower_violations = _immutable_copy(self.unscaled_lower_violations)
        self.unscaled_upper_values = _immutable_copy(self.unscaled_upper_values)
        self.unscaled_upper_violations = _immutable_copy(self.unscaled_upper_violations)

    @classmethod
    def create(
        cls, config: EnOptConfig, evaluations: FunctionEvaluations
    ) -> Optional[BoundConstraints]:
        """Add constraint information.

        This factory function creates a `BoundConstraints` object with the correct
        information on bound constraints, based on the
        given evaluations.

        Args:
            config:      The ensemble optimizer configuration object
            evaluations: An instance of the Evaluations class

        Returns:
            A newly created BoundConstraints object.
        """

        def _get_unscaled(
            scaled: Optional[NDArray[np.float64]],
        ) -> Optional[NDArray[np.float64]]:
            unscaled = None
            if scaled is not None:
                assert config is not None
                unscaled = utils.scaling.scale_back_variables(
                    config,
                    scaled,
                    correct_offsets=False,
                    axis=-1,
                )
                if unscaled is not None:
                    unscaled.setflags(write=False)
            return unscaled

        kwargs = {}
        if np.any(np.isfinite(config.variables.lower_bounds)):
            lower_values = _get_lower_bound_constraint_values(
                config, evaluations.variables, axis=-1
            )
            lower_violations = _get_lower_bound_violations(lower_values)
            kwargs.update(
                {
                    "lower_values": lower_values,
                    "lower_violations": lower_violations,
                    "unscaled_lower_values": _get_unscaled(lower_values),
                    "unscaled_lower_violations": _get_unscaled(lower_violations),
                },
            )
        if np.any(np.isfinite(config.variables.upper_bounds)):
            upper_values = _get_upper_bound_constraint_values(
                config, evaluations.variables, axis=-1
            )
            upper_violations = _get_upper_bound_violations(upper_values)
            kwargs.update(
                {
                    "upper_values": upper_values,
                    "upper_violations": upper_violations,
                    "unscaled_upper_values": _get_unscaled(upper_values),
                    "unscaled_upper_violations": _get_unscaled(upper_violations),
                },
            )
        return BoundConstraints(**kwargs) if kwargs else None



import pytest

pytest.importorskip("parsl")

import os
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, cast, no_type_check

import numpy as np
from numpy.typing import NDArray
from parsl.app.app import python_app

from ropt.evaluator import EvaluatorContext
from ropt.evaluator.parsl import ParslEvaluator, State, Task
from ropt.plan import OptimizationPlanRunner


@dataclass
class ParslTestTask(Task):
    def get_objectives(self) -> Optional[NDArray[np.float64]]:
        assert self.future is not None
        result = self.future.result()
        if result is None:
            return None
        return cast(NDArray[np.float64], result)


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
        },
        "optimizer": {
            "max_functions": 3,
        },
        "objective_functions": {
            "weights": [0.75, 0.25],
        },
        "realizations": {
            "weights": [1.0] * 3,
            "realization_min_success": 1,
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
            "perturbation_min_success": 2,
        },
    }


@python_app
@no_type_check
def _run_functions(
    functions: Tuple[Callable[..., Any], ...],
    variables: NDArray[np.float64],
    fail: bool = False,
) -> NDArray[np.float64]:
    result = np.zeros(len(functions), dtype=np.float64)
    for idx, function in enumerate(functions):
        if fail:
            raise RuntimeError
        result[idx] = function(variables, None)
    return result


def parsl_function(
    batch_id: int,  # noqa: ARG001
    idx: int,
    variables: NDArray[np.float64],
    context: EvaluatorContext,  # noqa: ARG001
    functions: Tuple[Callable[..., Any], ...],
    fail_index: int = -1,
) -> List[ParslTestTask]:
    return [
        ParslTestTask(
            future=_run_functions(functions, variables, fail=idx == fail_index)
        ),
    ]


def parsl_monitor(batch_id: int, jobs: Dict[int, List[ParslTestTask]]) -> None:
    for job_idx, tasks in jobs.items():
        for task_idx, task in enumerate(tasks):
            if task.state == State.FAILED:
                print(f"error in job {job_idx}")  # noqa: T201
            if task.state == State.SUCCESS:
                print(  # noqa: T201
                    f"batch: {batch_id}, job: {job_idx}, "
                    f"task: {task_idx} has finished"
                )


def test_parsl(enopt_config: Any, test_functions: Any, tmp_path: Any) -> None:
    os.chdir(tmp_path)
    with ParslEvaluator(
        function=partial(parsl_function, functions=test_functions)
    ) as evaluator:
        variables = OptimizationPlanRunner(enopt_config, evaluator).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)


def test_parsl_dummy_htex(
    enopt_config: Any, test_functions: Any, tmp_path: Any
) -> None:
    os.chdir(tmp_path)
    with ParslEvaluator(
        function=partial(parsl_function, functions=test_functions)
    ).with_htex(provider=None) as evaluator:
        variables = OptimizationPlanRunner(enopt_config, evaluator).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)


def test_parsl_monitor(
    enopt_config: Any, test_functions: Any, tmp_path: Any, capsys: Any
) -> None:
    os.chdir(tmp_path)
    with ParslEvaluator(
        function=partial(parsl_function, functions=test_functions)
    ).with_monitor(
        parsl_monitor,
    ) as evaluator:
        variables = OptimizationPlanRunner(enopt_config, evaluator).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)
    captured = capsys.readouterr()
    assert "batch: 0, job: 0, task: 0 has finished" in captured.out


def test_parsl_exception(
    enopt_config: Any, test_functions: Any, tmp_path: Any, capsys: Any
) -> None:
    os.chdir(tmp_path)
    evaluator = ParslEvaluator(
        function=partial(parsl_function, functions=test_functions, fail_index=2),
    ).with_monitor(
        parsl_monitor,
    )
    with evaluator:
        variables = OptimizationPlanRunner(enopt_config, evaluator).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)
    captured = capsys.readouterr()
    assert "error in job 2" in captured.out


def test_parsl_no_with(enopt_config: Any, test_functions: Any, tmp_path: Any) -> None:
    os.chdir(tmp_path)
    evaluator = ParslEvaluator(
        function=partial(parsl_function, functions=test_functions, fail_index=2),
    )
    with pytest.raises(RuntimeError):
        OptimizationPlanRunner(enopt_config, evaluator).run()



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

from ropt import utils
from ropt.enums import ResultAxisName

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig


@dataclass
class GradientEvaluations(ResultField):
    """This class contains the results of evaluations for gradient calculation.

    This class stores the variables with the calculated objective and
    constraints gradients. It contains the following information:

    1. The vector of variables at which the functions are evaluated.
    2. A three-dimensional array of perturbed variables, with variable values
       arranged along the third axis. The second axis index indicates the
       perturbation number, whereas the first axis index represents the
       realization number.
    3. The objectives and constraints for each realization and perturbed
       variable vector:  A three-dimensional array, with the objective or
       constraint values arranged along the third axis. The second axis index
       indicates the perturbation number, whereas the first axis index
       represents the realization number.
    4. Scaled versions of the objective and constraint values if scaling was
       enabled.
    5. Optional evaluation IDs that may have been passed from the evaluator,
       identifying each calculated realization and perturbation.

    Args:
        variables:                    The unperturbed variable vector.
        perturbed_variables:          The variables for each realization and perturbation.
        perturbed_objectives:         The objective functions for each realization and
                                      perturbation.
        perturbed_constraints:        The constraint functions for each realization and
                                      perturbation.
        unscaled_variables:           Optional variables after scaling back.
        unscaled_perturbed_variables: Optional variables after scaling back.
        scaled_perturbed_objectives:  Optional scaled objectives.
        scaled_perturbed_constraints: Optional scaled constraints.
        perturbed_evaluation_ids:     Optional id of each evaluated realization.
    """

    variables: NDArray[np.float64] = field(
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    perturbed_variables: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.VARIABLE,
            ),
        },
    )
    perturbed_objectives: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.OBJECTIVE,
            ),
        },
    )
    perturbed_constraints: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.NONLINEAR_CONSTRAINT,
            ),
        },
    )
    unscaled_variables: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.VARIABLE,),
        },
    )
    unscaled_perturbed_variables: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.VARIABLE,
            ),
        },
    )
    scaled_perturbed_objectives: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.OBJECTIVE,
            ),
        },
    )
    scaled_perturbed_constraints: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
                ResultAxisName.NONLINEAR_CONSTRAINT,
            ),
        },
    )
    perturbed_evaluation_ids: Optional[NDArray[np.intc]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.REALIZATION,
                ResultAxisName.PERTURBATION,
            ),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.variables = _immutable_copy(self.variables)
        self.perturbed_variables = _immutable_copy(self.perturbed_variables)
        self.perturbed_objectives = _immutable_copy(self.perturbed_objectives)
        self.perturbed_constraints = _immutable_copy(self.perturbed_constraints)
        self.unscaled_variables = _immutable_copy(self.unscaled_variables)
        self.unscaled_perturbed_variables = _immutable_copy(
            self.unscaled_perturbed_variables
        )
        self.scaled_perturbed_objectives = _immutable_copy(
            self.scaled_perturbed_objectives
        )
        self.scaled_perturbed_constraints = _immutable_copy(
            self.scaled_perturbed_constraints
        )
        self.perturbed_evaluation_ids = _immutable_copy(self.perturbed_evaluation_ids)

    @classmethod
    def create(  # noqa: PLR0913
        cls,
        config: EnOptConfig,
        variables: NDArray[np.float64],
        perturbed_variables: NDArray[np.float64],
        perturbed_objectives: NDArray[np.float64],
        perturbed_constraints: Optional[NDArray[np.float64]] = None,
        objective_auto_scales: Optional[NDArray[np.float64]] = None,
        constraint_auto_scales: Optional[NDArray[np.float64]] = None,
        perturbed_evaluation_ids: Optional[NDArray[np.intc]] = None,
    ) -> GradientEvaluations:
        """Create a FunctionEvaluations object with the given information.

        Args:
            config:                   Configuration object
            variables:                The unperturbed variable vector
            perturbed_variables:      The unperturbed variable vector
            perturbed_objectives:     The objective functions for each realization
            perturbed_constraints:    The constraint functions for each realization
            objective_auto_scales:    Objective auto-scaling information
            constraint_auto_scales:   Constraint auto-scaling information
            perturbed_evaluation_ids: Optional IDs of the objective calculations

        Returns:
            A new FunctionEvaluations object.
        """
        unscaled_variables = utils.scaling.scale_back_variables(
            config, variables, axis=-1
        )
        unscaled_perturbed_variables = utils.scaling.scale_back_variables(
            config, perturbed_variables, axis=-1
        )
        scaled_perturbed_objectives = utils.scaling.scale_objectives(
            config,
            perturbed_objectives,
            None if objective_auto_scales is None else objective_auto_scales,
            axis=-1,
        )
        scaled_perturbed_constraints = utils.scaling.scale_constraints(
            config,
            perturbed_constraints,
            None if constraint_auto_scales is None else constraint_auto_scales,
            axis=-1,
        )
        return GradientEvaluations(
            variables=variables,
            perturbed_variables=perturbed_variables,
            perturbed_objectives=perturbed_objectives,
            perturbed_constraints=perturbed_constraints,
            unscaled_variables=unscaled_variables,
            unscaled_perturbed_variables=unscaled_perturbed_variables,
            scaled_perturbed_objectives=scaled_perturbed_objectives,
            scaled_perturbed_constraints=scaled_perturbed_constraints,
            perturbed_evaluation_ids=perturbed_evaluation_ids,
        )



"""Functions for scaling variables and functions."""

from typing import Optional

import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig


def scale_variables(
    config: EnOptConfig, variables: NDArray[np.float64], axis: int
) -> Optional[NDArray[np.float64]]:
    """Scale Back Variables.

    Given an ensemble optimizer configuration object and a vector of variables,
    this function applies an offset and scale to their values.

    As the `variables` input may be a multi-dimensional array, the index of the
    axis that designates the variables should be provided through the `axis`
    argument.

    Args:
        config:    The ensemble optimizer configuration object.
        variables: The scaled variables.
        axis:      The variables axis.

    Returns:
        The unscaled variables, or `None` if no scaling is applied.
    """
    if config.variables.offsets is None and config.variables.scales is None:
        return None
    variables = np.moveaxis(variables, axis, -1)
    if config.variables.offsets is not None:
        variables = variables - config.variables.offsets
    if config.variables.scales is not None:
        variables = variables / config.variables.scales
    return np.moveaxis(variables, -1, axis)


def scale_back_variables(
    config: EnOptConfig,
    variables: NDArray[np.float64],
    axis: int,
    *,
    correct_offsets: bool = True,
) -> Optional[NDArray[np.float64]]:
    """Scale back variables.

    Given an ensemble optimizer configuration object and a vector of scaled
    variables, scale their values back to the original range. Normally this
    includes correcting for offsets, but if a difference value is being
    rescaled, the `correct_offsets` flag can be used to disable this.

    As the `variables` input may be a multi-dimensional array, the index of the
    axis that designates the variables should be provided through the `axis`
    argument.

    Args:
        config:          The ensemble optimizer configuration object
        variables:       The scaled variables
        axis:            The variables axis
        correct_offsets: If True also correct for offsets.

    Returns:
        The unscaled variables, or `None` if no scaling is applied.
    """
    if config.variables.offsets is None and config.variables.scales is None:
        return None
    variables = np.moveaxis(variables, axis, -1)
    if config.variables.scales is not None:
        variables = variables * config.variables.scales
    if correct_offsets and config.variables.offsets is not None:
        variables = variables + config.variables.offsets
    return np.moveaxis(variables, -1, axis)


def scale_objectives(
    config: EnOptConfig,
    objectives: NDArray[np.float64],
    scales: Optional[NDArray[np.float64]],
    axis: int,
) -> Optional[NDArray[np.float64]]:
    """Scale Objective Functions.

    Given an ensemble optimizer configuration object, this function scales the
    provided objective values. It divides them by the scale values given in the
    configuration object (if not `None`), and optionally also by the values
    given in the `scales` argument.

    As the `objectives` input may be a multi-dimensional array, the index of the
    axis that designates the objectives should be provided through the `axis`
    argument.

    Args:
        config:     The ensemble optimizer configuration object.
        objectives: Objective functions.
        scales:     Optional additional scales.
        axis:       The objectives axis.

    Returns:
        The scaled objectives or `None` if no scaling was applied.
    """
    total_scales = config.objective_functions.scales
    if scales is not None:
        total_scales = total_scales * scales
    if np.allclose(total_scales, 1.0, rtol=0.0, atol=1e-10):
        return None
    objectives = np.moveaxis(objectives, axis, -1)
    return np.moveaxis(objectives / total_scales, -1, axis)


def scale_constraints(
    config: EnOptConfig,
    constraints: Optional[NDArray[np.float64]],
    scales: Optional[NDArray[np.float64]],
    axis: int,
) -> Optional[NDArray[np.float64]]:
    """Scale constraint functions.

    Given an ensemble optimizer configuration object, this function scales the
    provided constraint values. It divides them by the scale values given in the
    configuration object (if not `None`), and optionally also by the values
    given in the `scales` argument.

    As the `constraints` input may be a multi-dimensional array, the index of the
    axis that designates the constraints should be provided through the `axis`
    argument.

    Args:
        config:      The ensemble optimizer configuration object
        constraints: Constraint functions
        scales:      Optional additional scales.
        axis:        The constraints axis

    Returns:
        The scaled constraints or `None` if no scaling was applied.
    """
    if constraints is None or config.nonlinear_constraints is None:
        return None
    total_scales = config.nonlinear_constraints.scales
    if scales is not None:
        total_scales = total_scales * scales
    if np.allclose(total_scales, 1.0, rtol=0.0, atol=1e-10):
        return None
    constraints = np.moveaxis(constraints, axis, -1)
    return np.moveaxis(constraints / total_scales, -1, axis)



r"""Code to run optimization plans."""

from ropt.optimizer import EnsembleOptimizer

from ._plan import MetaDataType, OptimizerContext, Plan
from ._run import OptimizationPlanRunner
from ._update import (
    ContextUpdate,
    ContextUpdateDict,
    ContextUpdateResults,
)

__all__ = [
    "OptimizationPlanRunner",
    "ContextUpdate",
    "ContextUpdateDict",
    "ContextUpdateResults",
    "EnsembleOptimizer",
    "OptimizerContext",
    "MetaDataType",
    "Plan",
]



from dataclasses import replace
from typing import Any, Dict

import numpy as np
import pytest

from ropt.config.enopt import EnOptConfig
from ropt.enums import ConstraintType
from ropt.results import (
    BoundConstraints,
    FunctionEvaluations,
    FunctionResults,
    Functions,
    GradientEvaluations,
    GradientResults,
    Gradients,
    LinearConstraints,
    NonlinearConstraints,
    Realizations,
)


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "variables": {
            "names": ["x", "y"],
            "initial_values": [0.0, 0.0],
        },
        "objective_functions": {
            "names": ["f1", "f2"],
            "weights": [0.75, 0.25],
        },
        "realizations": {
            "names": ["r1", "r2", "r3"],
            "weights": [1.0] * 3,
        },
        "gradient": {
            "number_of_perturbations": 5,
        },
    }


@pytest.fixture(name="function_result")
def function_result_fixture(enopt_config: Any) -> FunctionResults:
    config = EnOptConfig.model_validate(enopt_config)
    evaluations = FunctionEvaluations.create(
        config=config,
        variables=np.array([1.0, 2.0]),
        objectives=np.arange(6, dtype=np.float64).reshape((3, 2)),
    )
    realizations = Realizations(
        objective_weights=np.arange(6, dtype=np.float64).reshape((2, 3)),
        failed_realizations=np.zeros(3, dtype=np.bool_),
    )
    functions = Functions.create(
        config=config, weighted_objective=np.array(1.0), objectives=np.array([1.0, 2.0])
    )
    return FunctionResults(
        result_id=0,
        batch_id=1,
        metadata={},
        evaluations=evaluations,
        realizations=realizations,
        functions=functions,
    )


@pytest.fixture(name="gradient_result")
def gradient_result_fixture() -> GradientResults:
    evaluations = GradientEvaluations(
        variables=np.array([1.0, 2.0]),
        perturbed_variables=np.arange(30, dtype=np.float64).reshape((3, 5, 2)),
        perturbed_objectives=np.arange(30, dtype=np.float64).reshape((3, 5, 2)),
    )
    gradients = Gradients(
        weighted_objective=np.array([1.0, 2.0]),
        objectives=np.arange(4, dtype=np.float64).reshape((2, 2)),
    )
    return GradientResults(
        result_id=0,
        batch_id=1,
        metadata={},
        evaluations=evaluations,
        realizations=Realizations(
            failed_realizations=np.zeros(36, dtype=np.bool_),
        ),
        gradients=gradients,
    )


def test_get_axis_names_error(function_result: FunctionResults) -> None:
    with pytest.raises(ValueError, match="Unknown field name: foo"):
        function_result.evaluations.get_axis_names("foo")


def test_scaling_evaluations_functions(
    enopt_config: Any, function_result: FunctionResults
) -> None:
    enopt_config["variables"]["scales"] = [2.0, 2.0]
    enopt_config["objective_functions"]["scales"] = [2.0, 2.0]
    config = EnOptConfig.model_validate(enopt_config)
    evaluations = FunctionEvaluations.create(
        config=config,
        variables=np.array([1.0, 2.0]),
        objectives=np.arange(6, dtype=np.float64).reshape((3, 2)),
    )
    function_result = replace(function_result, evaluations=evaluations)
    assert config.variables.scales is not None
    assert function_result.evaluations.unscaled_variables is not None
    assert np.allclose(
        function_result.evaluations.unscaled_variables,
        function_result.evaluations.variables * config.variables.scales,
    )
    assert function_result.evaluations.scaled_objectives is not None
    assert np.allclose(
        function_result.evaluations.scaled_objectives,
        function_result.evaluations.objectives / config.objective_functions.scales,
    )


def test_scaling_evaluations_functions_auto(
    enopt_config: Any, function_result: FunctionResults
) -> None:
    enopt_config["objective_functions"]["scales"] = [2.0, 2.0]
    config = EnOptConfig.model_validate(enopt_config)
    objective_auto_scales = np.array([3.0, 3.0])
    evaluations = FunctionEvaluations.create(
        config=config,
        objective_auto_scales=objective_auto_scales,
        variables=np.array([1.0, 2.0]),
        objectives=np.arange(6, dtype=np.float64).reshape((3, 2)),
    )
    function_result = replace(function_result, evaluations=evaluations)
    assert function_result.evaluations.scaled_objectives is not None
    assert np.allclose(
        function_result.evaluations.scaled_objectives,
        function_result.evaluations.objectives
        / config.objective_functions.scales
        / objective_auto_scales,
    )


def test_scaling_functions(enopt_config: Any, function_result: FunctionResults) -> None:
    enopt_config["objective_functions"]["scales"] = [2.0, 2.0]
    config = EnOptConfig.model_validate(enopt_config)
    functions = Functions.create(
        config=config,
        weighted_objective=np.array(1.0),
        objectives=np.array([1.0, 2.0]),
    )
    function_result = replace(function_result, functions=functions)
    assert function_result.functions is not None
    assert function_result.functions.scaled_objectives is not None
    assert np.allclose(
        function_result.functions.scaled_objectives,
        function_result.functions.objectives / config.objective_functions.scales,
    )


def test_scaling_functions_auto(
    enopt_config: Any, function_result: FunctionResults
) -> None:
    enopt_config["objective_functions"]["scales"] = [2.0, 2.0]
    config = EnOptConfig.model_validate(enopt_config)
    objective_auto_scales = np.array([3.0, 3.0])
    functions = Functions.create(
        config=config,
        objective_auto_scales=objective_auto_scales,
        weighted_objective=np.array(1.0),
        objectives=np.array([1.0, 2.0]),
    )
    function_result = replace(function_result, functions=functions)
    assert function_result.functions is not None
    assert function_result.functions.scaled_objectives is not None
    assert np.allclose(
        function_result.functions.scaled_objectives,
        function_result.functions.objectives
        / config.objective_functions.scales
        / objective_auto_scales,
    )


def test_bound_constraint_results(
    enopt_config: Any, function_result: FunctionResults
) -> None:
    enopt_config["variables"]["lower_bounds"] = [1.2, 1.4]
    enopt_config["variables"]["upper_bounds"] = [1.0, 1.5]
    config = EnOptConfig.model_validate(enopt_config)
    assert function_result.functions is not None
    constraints = BoundConstraints.create(config, function_result.evaluations)
    assert constraints is not None
    assert constraints.lower_values is not None
    assert constraints.upper_values is not None
    assert constraints.lower_violations is not None
    assert constraints.upper_violations is not None
    assert np.allclose(constraints.lower_values, [-0.2, 0.6])
    assert np.allclose(constraints.upper_values, [0.0, 0.5])
    assert np.allclose(constraints.lower_violations, [0.2, 0.0])
    assert np.allclose(constraints.upper_violations, [0.0, 0.5])


def test_linear_constraint_results(
    enopt_config: Any, function_result: FunctionResults
) -> None:
    enopt_config["linear_constraints"] = {
        "coefficients": [[1.0, 0.0], [0.0, 1.0]],
        "rhs_values": [0.0, 1.0],
        "types": [ConstraintType.LE, ConstraintType.GE],
    }
    config = EnOptConfig.model_validate(enopt_config)
    assert function_result.functions is not None
    constraints = LinearConstraints.create(config, function_result.evaluations)
    assert constraints is not None
    assert constraints.values is not None  # noqa: PD011
    assert constraints.violations is not None
    assert np.allclose(constraints.values, [1.0, 1.0])
    assert np.allclose(constraints.violations, [1.0, 0.0])


def test_nonlinear_constraint_results(
    enopt_config: Any, function_result: FunctionResults
) -> None:
    enopt_config["nonlinear_constraints"] = {
        "rhs_values": [0.0, 1.0, 0.0],
        "types": [ConstraintType.LE, ConstraintType.GE, ConstraintType.EQ],
    }
    config = EnOptConfig.model_validate(enopt_config)
    assert function_result.functions is not None
    function_result.functions.constraints = np.array([1.0, 1.0, -1.0])
    constraints = NonlinearConstraints.create(config, function_result.functions, None)
    assert constraints is not None
    assert constraints.values is not None  # noqa: PD011
    assert constraints.violations is not None
    assert np.allclose(constraints.values, [1.0, 0.0, -1.0])
    assert np.allclose(constraints.violations, [1.0, 0.0, 1.0])



"""This module implements the default optimizer step."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import BaseModel, ConfigDict

from ropt.exceptions import PlanError
from ropt.plugins.plan.base import PlanStep

if TYPE_CHECKING:
    from ropt.config.plan import StepConfig
    from ropt.plan import Plan


class DefaultResetStepWith(BaseModel):
    """Parameters used by the default reset step."""

    context: str

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
    )


class DefaultResetStep(PlanStep):
    """The default reset step."""

    def __init__(self, config: StepConfig, plan: Plan) -> None:
        """Initialize a default reset context step.

        Args:
            config: The configuration of the step
            plan:   The plan that runs this step
        """
        super().__init__(config, plan)

        self._context = (
            config.with_
            if isinstance(config.with_, str)
            else DefaultResetStepWith.model_validate(config.with_).context
        )

    def run(self) -> bool:
        """Run the reset step.

        Returns:
            True if a user abort occurred, always `False`.
        """
        if not self.plan.has_context(self._context):
            msg = f"Env object `{self._context}` does not exist."
            raise PlanError(msg, step_name=self.step_config.name)
        self.plan.reset_context(self._context)
        return False



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

from ropt.enums import ResultAxisName

from ._result_field import ResultField
from ._utils import _immutable_copy

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray


@dataclass
class Gradients(ResultField):
    """This class stores the calculated objective and constraint gradients.

    The objective and constraint gradients used by the optimizer are calculated
    from their values evaluated for all realizations, for instance by averaging.
    There may be multiple objectives and constraints. Multiple objectives are
    handled by the optimizer using a weighted sum of the individual gradients,
    stored in the `weighted_objective` field. Multiple constraints are directly
    handled by the optimizer.

    Args:
        weighted_objective: The weighted sum of the objective gradients
        objectives:         The value of each objective gradient
        constraints:        The value of each constraint gradient
    """

    weighted_objective: NDArray[np.float64] = field(
        metadata={"__axes__": (ResultAxisName.VARIABLE,)},
    )
    objectives: NDArray[np.float64] = field(
        metadata={
            "__axes__": (
                ResultAxisName.OBJECTIVE,
                ResultAxisName.VARIABLE,
            ),
        },
    )
    constraints: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (
                ResultAxisName.NONLINEAR_CONSTRAINT,
                ResultAxisName.VARIABLE,
            ),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.weighted_objective = _immutable_copy(self.weighted_objective)
        self.objectives = _immutable_copy(self.objectives)
        self.constraints = _immutable_copy(self.constraints)



from typing import Any, Dict

import numpy as np
import pytest

from ropt.config.enopt import EnOptConfig
from ropt.results import FunctionEvaluations, FunctionResults, Functions, Realizations

pytest.importorskip("xarray")


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "variables": {
            "names": ["x", "y"],
            "initial_values": [0.0, 0.0],
        },
        "objective_functions": {
            "names": ["f1", "f2"],
            "weights": [0.75, 0.25],
        },
        "realizations": {
            "names": ["r1", "r2", "r3"],
            "weights": [1.0] * 3,
        },
        "gradient": {
            "number_of_perturbations": 5,
        },
    }


@pytest.fixture(name="function_result")
def function_result_fixture(enopt_config: Any) -> FunctionResults:
    config = EnOptConfig.model_validate(enopt_config)
    evaluations = FunctionEvaluations.create(
        config=config,
        variables=np.array([1.0, 2.0]),
        objectives=np.arange(6, dtype=np.float64).reshape((3, 2)),
    )
    realizations = Realizations(
        objective_weights=np.arange(6, dtype=np.float64).reshape((2, 3)),
        failed_realizations=np.zeros(3, dtype=np.bool_),
    )
    functions = Functions.create(
        config=config,
        weighted_objective=np.array(1.0),
        objectives=np.array([1.0, 2.0]),
    )
    return FunctionResults(
        result_id=0,
        batch_id=1,
        metadata={"foo": "bar"},
        evaluations=evaluations,
        realizations=realizations,
        functions=functions,
    )


def test_to_dataset(enopt_config: Any, function_result: FunctionResults) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    dataset = function_result.to_dataset(config, "evaluations")
    for field in dataset:
        assert np.all(
            dataset[field].to_numpy()
            == getattr(function_result.evaluations, str(field)),
        )
    assert (
        tuple(dataset.coords["variable"].values)
        == config.variables.get_formatted_names()
    )
    assert tuple(dataset.coords["realization"].values) == config.realizations.names
    assert tuple(dataset.coords["objective"].values) == config.objective_functions.names


@pytest.mark.parametrize("add_metadata", [True, False])
def test_to_dataset_with_attrs(
    enopt_config: Any, function_result: FunctionResults, add_metadata: bool
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    dataset = function_result.to_dataset(
        config, "evaluations", add_metadata=add_metadata
    )
    assert dataset.attrs["result_id"] == 0
    assert dataset.attrs["batch_id"] == 1
    if add_metadata:
        assert dataset.attrs["metadata"] == {"foo": "bar"}
    else:
        assert "metadata" not in dataset.attrs


def test_to_dataset_formatter(
    enopt_config: Any, function_result: FunctionResults
) -> None:
    enopt_config["variables"]["names"] = [("x", 0), ("x", 1)]
    config = EnOptConfig.model_validate(enopt_config)
    dataset = function_result.to_dataset(config, "evaluations", select=["variables"])
    assert dataset.coords["variable"].to_numpy().tolist() == ["x:0", "x:1"]



import copy
from functools import singledispatch

from ._function_results import FunctionResults
from ._gradient_results import GradientResults
from ._results import Results


@singledispatch
def convert_to_maximize(results: Results) -> Results:
    """Convert results to maximization results.

    The optimization process always assumes that the objective functions must be
    minimized. To perform a maximization, the negative of the objective
    functions can be passed instead. However, the result objects will then
    contain those negative values and all objective function values derived from
    it. Since this may be inconvenient for the application, this convenience
    function can be used make a copy of a result where all objective values are
    multiplied by -1, which would correspond to a maximization.

    Args:
        results: The result to convert

    Returns:
        A copy of the result, modified to reflect maximization.
    """
    msg = f"`results` type not supported: {type(results).__name__}"
    raise TypeError(msg)


@convert_to_maximize.register
def _(results: FunctionResults) -> FunctionResults:
    results = copy.deepcopy(results)
    results.evaluations.objectives.setflags(write=True)
    results.evaluations.objectives *= -1.0
    results.evaluations.objectives.setflags(write=False)
    if results.evaluations.scaled_objectives is not None:
        results.evaluations.scaled_objectives.setflags(write=True)
        results.evaluations.scaled_objectives *= -1.0
        results.evaluations.scaled_objectives.setflags(write=False)
    if results.functions is not None:
        results.functions.weighted_objective *= -1.0
        results.functions.objectives.setflags(write=True)
        results.functions.objectives *= -1.0
        results.functions.objectives.setflags(write=False)
        if results.functions.scaled_objectives is not None:
            results.functions.scaled_objectives.setflags(write=True)
            results.functions.scaled_objectives *= -1.0
            results.functions.scaled_objectives.setflags(write=False)
    return results


@convert_to_maximize.register
def _(results: GradientResults) -> GradientResults:
    results = copy.deepcopy(results)
    results.evaluations.perturbed_objectives.setflags(write=True)
    results.evaluations.perturbed_objectives *= -1.0
    results.evaluations.perturbed_objectives.setflags(write=False)
    if results.gradients is not None:
        results.gradients.weighted_objective *= -1.0
        results.gradients.objectives.setflags(write=True)
        results.gradients.objectives *= -1.0
        results.gradients.objectives.setflags(write=False)
    return results



"""Simple application object to run script-based optimization plans with parsl."""

from ._config import ScriptOptimizerConfig
from ._script_optimizer import ScriptOptimizer

__all__ = [
    "ScriptOptimizer",
    "ScriptOptimizerConfig",
]



"""Miscellaneous utilities."""

from copy import deepcopy
from typing import Any, Dict


def update_dict(mapping: Dict[str, Any], values: Dict[str, Any]) -> Dict[str, Any]:
    """Recursively update a dictionary.

    Args:
        mapping: The dictionary to update
        values:  The values to be updated or added

    Returns:
        The updated dictionary.
    """
    mapping = mapping.copy()
    for key, value in values.items():
        if (
            key in mapping
            and isinstance(mapping[key], dict)
            and isinstance(value, dict)
        ):
            mapping[key] = update_dict(mapping[key], value)
        else:
            mapping[key] = deepcopy(value)
    return mapping



"""This module implements the default setvar step."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from pydantic import BaseModel, ConfigDict

from ropt.exceptions import PlanError
from ropt.plugins.plan.base import PlanStep

if TYPE_CHECKING:
    from ropt.config.plan import StepConfig
    from ropt.plan import Plan


class DefaultSetStepExprWith(BaseModel):
    """Parameters used by the default setvar step.

    Attributes:
        expr: Expression used to set the variable
    """

    expr: str

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
        arbitrary_types_allowed=True,
    )


class DefaultSetStepVarValueWith(BaseModel):
    """Parameters used by the default setvar step.

    Attributes:
        var:   The variable to set
        value: The value
    """

    var: str
    value: Any

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
        arbitrary_types_allowed=True,
    )


class DefaultSetStep(PlanStep):
    """The default set step."""

    def __init__(self, config: StepConfig, plan: Plan) -> None:
        """Initialize a default setvar step.

        Args:
            config: The configuration of the step
            plan:   The plan that runs this step
        """
        super().__init__(config, plan)

        expr: Optional[str] = None
        var: str
        value: Any

        if isinstance(config.with_, str):
            expr = config.with_
        elif "expr" in config.with_:
            expr = DefaultSetStepExprWith.model_validate(config.with_).expr
        elif "var" in config.with_:
            with_ = DefaultSetStepVarValueWith.model_validate(config.with_)
            var = with_.var
            value = with_.value
        else:
            msg = "Either `expr` or `var` must be provided"
            raise RuntimeError(msg)

        if expr is not None:
            var, sep, value = expr.partition("=")
            if sep != "=":
                msg = f"Invalid expression: {expr}"
                raise PlanError(msg, step_name=self._step_config.name)

        self._var = var.strip()
        if not self._var.isidentifier():
            msg = f"Invalid identifier: {self._var}"
            raise PlanError(msg, step_name=self._step_config.name)
        self._value = value

    def run(self) -> bool:
        """Run the setvar step.

        Returns:
            True if a user abort occurred, always `False`.
        """
        self._plan[self._var] = self._plan.eval(self._value)
        return False



from __future__ import annotations

from dataclasses import dataclass, fields
from typing import TYPE_CHECKING, Tuple, TypeVar

if TYPE_CHECKING:
    from ropt.enums import ResultAxisName


TypeResultField = TypeVar("TypeResultField", bound="ResultField")


@dataclass
class ResultField:
    """Base class for `Results` fields."""

    @classmethod
    def get_axis_names(cls, name: str) -> Tuple[ResultAxisName, ...]:
        """Return the axis names of a field in the given field class or object.

        When used with the class or an instance of that class for one of the
        fields of a `Results` object, retrieve the names of the axes of the
        stored `numpy` array from the metadata and return them.

        Args:
            name: The name of the sub-field in the instance or class

        Raises:
            ValueError: Raised if an unknown field name is passed.

        Returns:
            A tuple listing the names of the axes.
        """
        metadata = next(
            (item.metadata for item in fields(cls) if item.name == name), None
        )
        if metadata is None:
            msg = f"Unknown field name: {name}"
            raise ValueError(msg)
        return metadata.get("__axes__", ())



"""The `plugins` module facilitates the integration of `ropt` plugins.

The core functionality of `ropt` can be extended through plugins, which can
either be builtin, separately installed, or dynamically added at runtime.
Currently, `ropt` supports the following plugin types to implement specific
types of functionality:

[`optimizer`][ropt.plugins.optimizer]:
: Plugins that implement specific optimization methods. The builtin
  [`scipy`][ropt.plugins.optimizer.scipy.SciPyOptimizer] plugin utilizes the
  [`scipy.optimize`](https://docs.scipy.org/doc/scipy/reference/optimize.html)
module to implement various optimization algorithms.

[`sampler`][ropt.plugins.sampler]:
: Plugins responsible for generating perturbations for estimating gradients.
  The builtin [`scipy`][ropt.plugins.sampler.scipy.SciPySampler] plugin is based
  on [`scipy.stats`](https://docs.scipy.org/doc/scipy/reference/stats.html),
  providing various sampling methods.

[`realization_filter`][ropt.plugins.realization_filter]:
: These plugins implement filters for selecting a sub-set of realizations used
  in calculating objective or constraint functions and their gradients. The
  included
  [`default`][ropt.plugins.realization_filter.default.DefaultRealizationFilter]
  plugin provides filters based on ranking and for CVaR optimization.

[`function_transform`][ropt.plugins.function_transform]:
: These plugins calculate the final objective and gradient from sets of
  objectives or constraints and their gradients for individual realizations. The
  included
  [`default`][ropt.plugins.function_transform.default.DefaultFunctionTransform]
  plugin supports objectives defined by the mean or standard deviation of these
  values.

[`plan`][ropt.plugins.plan]:
: Plan plugins implement the objects that execute an optimization plan.
  The built-in [`default`][ropt.plugins.plan.default.DefaultPlanPlugin] plugin
  offers a full set of optimization plan objects for executing complex
  optimization plans.

Plugins are managed by the [`PluginManager`][ropt.plugins.PluginManager] class.
This class is used to retrieve plugin objects that derive from an abstract
base class defining the required functionality for each plugin type:

1. [`OptimizerPlugin`][ropt.plugins.optimizer.base.OptimizerPlugin]:
    Abstract base class for optimizer plugins.
2. [`SamplerPlugin`][ropt.plugins.sampler.base.SamplerPlugin]:
    Abstract base class for sampler plugins.
3. [`RealizationFilterPlugin`][ropt.plugins.realization_filter.base.RealizationFilterPlugin]:
    Abstract base class for realization filter plugins.
4. [`FunctionTransformPlugin`][ropt.plugins.function_transform.base.FunctionTransformPlugin]:
    Abstract base class for function transform plugins.
5. [`PlanPlugin`][ropt.plugins.plan.base.PlanPlugin]:
    Abstract base class for optimization plan object plugins.

Plugins can be built-in, installed separately using the standard entry points
mechanism, or added dynamically using the
[`add_plugins`][ropt.plugins.PluginManager.add_plugins] method.

The plugin manager object provides the
[`get_plugin`][ropt.plugins.PluginManager.get_plugin] method, which `ropt` uses
to retrieve the necessary plugin based on its type and name. Given the plugin's
type and name, this method returns a callable (either a class or a factory
function) that `ropt` uses to instantiate the plugin when needed.
"""

from ._manager import PluginManager, PluginType
from .base import Plugin

__all__ = [
    "Plugin",
    "PluginManager",
    "PluginType",
]



"""Base model class for `EnOptConfig` fields."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class EnOptBaseModel(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        str_min_length=1,
        str_strip_whitespace=True,
        validate_default=True,
    )



from __future__ import annotations

from dataclasses import fields
from typing import TYPE_CHECKING, Any, Iterable, Optional, Tuple, cast

import pandas as pd

from ropt.enums import ResultAxisName

if TYPE_CHECKING:
    from ropt.config.enopt import EnOptConfig

    from ._result_field import ResultField


def _to_dataframe(  # noqa: PLR0913
    config: EnOptConfig,
    result_field: ResultField,
    result_id: int,
    batch_id: Optional[int],
    select: Optional[Iterable[str]],
    unstack: Optional[Iterable[ResultAxisName]],
) -> pd.DataFrame:
    if select is None:
        select = (field.name for field in fields(result_field))
    if unstack is None:
        unstack = []
    joined_frame = pd.DataFrame()
    for field in select:
        series = _to_series(config, result_field, result_id, batch_id, field)
        if series is not None:
            frame = series.to_frame()
            for axis in unstack:
                if axis.value in frame.index.names:
                    frame = cast(pd.DataFrame, frame.unstack(axis.value))  # noqa: PD010
            frame.columns = frame.columns.to_flat_index()  # type:ignore
            if joined_frame.empty:
                joined_frame = frame
            else:
                joined_frame = joined_frame.join(frame, how="inner")
    return joined_frame


def _to_series(
    config: EnOptConfig,
    result_field: ResultField,
    result_id: int,
    batch_id: Optional[int],
    field: str,
) -> Optional[pd.Series[Any]]:
    try:
        data = getattr(result_field, field)
    except AttributeError as exc:
        msg = f"Not a field name: {field}"
        raise ValueError(msg) from exc
    if data is None:
        return None
    axes = result_field.get_axis_names(field)
    indices = [_get_index(config, axis) for axis in axes]
    indices = [
        index if index else pd.RangeIndex(data.shape[idx])
        for idx, index in enumerate(indices)
    ]
    series: pd.Series[Any]
    index: Tuple[Any, ...] = (result_id, 0 if batch_id is None else batch_id)
    index_names = ["result_id", "batch_id"]
    if indices:
        multi_index = pd.MultiIndex.from_product(
            indices, names=(axis.value for axis in axes)
        )
        series = pd.Series(data.flatten(), index=multi_index, name=field)
        series = pd.concat({index: series}, names=index_names)
    else:
        series = pd.Series(
            data,
            index=pd.MultiIndex.from_tuples([index], names=index_names),
            name=field,
        )
    return series


def _get_index(config: EnOptConfig, axis: ResultAxisName) -> Iterable[Any]:
    result: Iterable[Any] = []
    if axis == ResultAxisName.VARIABLE:
        formatted_names = config.variables.get_formatted_names()
        result = [] if formatted_names is None else formatted_names
    elif axis == ResultAxisName.OBJECTIVE:
        result = (
            config.objective_functions.names
            if config.objective_functions.names is not None
            else []
        )
    elif axis == ResultAxisName.NONLINEAR_CONSTRAINT:
        assert config.nonlinear_constraints is not None
        result = (
            config.nonlinear_constraints.names
            if config.nonlinear_constraints.names is not None
            else []
        )
    elif axis == ResultAxisName.REALIZATION:
        result = (
            config.realizations.names if config.realizations.names is not None else []
        )
    return result



"""This module implements the default optimizer step."""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from pydantic import BaseModel, ConfigDict

from ropt.config.plan import StepConfig  # noqa: TCH001
from ropt.plugins.plan.base import PlanStep

if TYPE_CHECKING:
    from ropt.plan import Plan


class DefaultRepeatStepWith(BaseModel):
    """Parameters used by the default repeat step.

    Attributes:
        iterations:  The number of repetitions
        steps:       The steps to repeat
        counter_var: The variable to update with the counter value
    """

    iterations: int
    steps: List[StepConfig]
    counter_var: Optional[str] = None

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
    )


class DefaultRepeatStep(PlanStep):
    """The default optimizer step."""

    def __init__(self, config: StepConfig, plan: Plan) -> None:
        """Initialize a default optimizer step.

        Args:
            config: The configuration of the step
            plan:   The plan that runs this step
        """
        super().__init__(config, plan)

        with_ = DefaultRepeatStepWith.model_validate(config.with_)
        self._num = with_.iterations
        self._steps = self.plan.create_steps(with_.steps)
        self._counter_var = with_.counter_var

    def run(self) -> bool:
        """Run the steps repeatedly.

        Returns:
            True if a user abort occurred.
        """
        for idx in range(self._num):
            if self._counter_var is not None:
                self.plan[self._counter_var] = idx
            if self.plan.run_steps(self._steps):
                return True
        return False



"""This module defines a basic optimization object."""

from __future__ import annotations

import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    NoReturn,
    Optional,
    Tuple,
    Union,
)

from ropt.config.plan import PlanConfig
from ropt.enums import EventType, OptimizerExitCode
from ropt.exceptions import OptimizationAborted
from ropt.plugins import PluginManager

from ._plan import OptimizerContext, Plan

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig
    from ropt.evaluator import Evaluator
    from ropt.events import Event
    from ropt.plugins._manager import PluginType
    from ropt.plugins.base import Plugin
    from ropt.results import FunctionResults

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class OptimizationPlanRunner:
    """A class for running optimization plans."""

    def __init__(
        self,
        enopt_config: Union[Dict[str, Any], EnOptConfig],
        evaluator: Evaluator,
        *,
        constraint_tolerance: float = 1e-10,
        seed: Optional[int] = None,
    ) -> None:
        self._enopt_config = enopt_config
        self._evaluator = evaluator
        self._plugin_manager: Optional[PluginManager] = None
        self._context = OptimizerContext(evaluator=self._evaluator, seed=seed)
        self._results: Optional[FunctionResults]
        self._variables: Optional[NDArray[np.float64]]
        self._exit_code: OptimizerExitCode = OptimizerExitCode.UNKNOWN
        self._observers: List[Tuple[EventType, Callable[[Event], None]]] = []

        self._plan_config: Dict[str, List[Dict[str, Any]]] = {
            "context": [
                {
                    "id": "config",
                    "init": "config",
                    "with": enopt_config,
                },
                {
                    "id": "optimal",
                    "init": "tracker",
                    "with": {"constraint_tolerance": constraint_tolerance},
                },
            ],
            "steps": [
                {
                    "run": "optimizer",
                    "with": {
                        "config": "$config",
                        "update": ["optimal"],
                        "exit_code_var": "exit_code",
                    },
                },
            ],
        }

    @property
    def results(self) -> Optional[FunctionResults]:
        return self._results

    @property
    def variables(self) -> Optional[NDArray[np.float64]]:
        return self._variables

    @property
    def exit_code(self) -> OptimizerExitCode:
        return self._exit_code

    def add_plugins(self, plugin_type: PluginType, plugins: Dict[str, Plugin]) -> Self:
        if self._plugin_manager is None:
            self._plugin_manager = PluginManager()
            self._plugin_manager.add_plugins(plugin_type, plugins)
        return self

    def add_observer(
        self, event_type: EventType, function: Callable[[Event], None]
    ) -> Self:
        self._observers.append((event_type, function))
        return self

    def add_metadata(self, metadata: Dict[str, Any]) -> Self:
        steps = self._plan_config["steps"]
        idx = next(
            (idx for idx, step in enumerate(steps) if step["run"] == "repeat"), None
        )
        if idx is not None:
            steps = steps[idx]["with"]["steps"]
        idx = next(idx for idx, step in enumerate(steps) if step["run"] == "optimizer")
        steps[idx]["with"]["metadata"] = metadata
        return self

    def repeat(
        self,
        iterations: int,
        restart_from: Literal["initial", "last", "optimal", "last_optimal"] = "optimal",
        counter_var: Optional[str] = None,
    ) -> Self:
        if any(step["run"] == "repeat" for step in self._plan_config["steps"]):
            msg = "The repeat() method can only be called once."
            raise RuntimeError(msg)
        steps = self._plan_config["steps"]
        idx = next(idx for idx, step in enumerate(steps) if step["run"] == "optimizer")
        if restart_from in ["last", "last_optimal"]:
            self._plan_config["context"].append(
                {
                    "id": "repeat_tracker",
                    "init": "tracker",
                    "with": {"type": "last" if restart_from == "last" else "optimal"},
                }
            )
            steps[idx]["with"]["update"].append("repeat_tracker")
        if restart_from == "last":
            steps[idx]["with"]["initial_values"] = "$repeat_tracker"
        elif restart_from == "optimal":
            steps[idx]["with"]["initial_values"] = "$optimal"
        elif restart_from == "last_optimal":
            steps[idx]["with"]["initial_values"] = "$initial"
            steps = [
                {"run": "setvar", "with": "initial = $repeat_tracker"},
                {"run": "reset", "with": {"context": "repeat_tracker"}},
                *steps,
            ]
        self._plan_config["steps"] = [
            {
                "run": "repeat",
                "with": {
                    "counter_var": counter_var,
                    "iterations": iterations,
                    "steps": steps,
                },
            }
        ]
        return self

    def run(self) -> Self:
        config = PlanConfig.model_validate(self._plan_config)
        plan = Plan(
            config,
            self._context,
            plugin_manager=self._plugin_manager,
        )
        for event_type, function in self._observers:
            plan.optimizer_context.events.add_observer(event_type, function)
        plan.run()
        self._results = plan["optimal"]
        self._variables = (
            None if self._results is None else self._results.evaluations.variables
        )
        self._exit_code = plan["exit_code"]
        return self

    @staticmethod
    def abort_optimization() -> NoReturn:
        """Abort the current optimization run.

        This method can be called from within callbacks to interrupt the ongoing
        optimization plan. The exact point at which the optimization is aborted
        depends on the step that is executing at that point. For example, within
        a running optimizer, the process will be interrupted after completing
        the current function evaluation.
        """
        raise OptimizationAborted(exit_code=OptimizerExitCode.USER_ABORT)



from functools import partial
from typing import Any, Dict

import pytest

from ropt.config.enopt import EnOptConfig
from ropt.enums import EventType
from ropt.events import Event
from ropt.plan import OptimizationPlanRunner
from ropt.report import ResultsDataFrame

# Requires pandas:
pd = pytest.importorskip("pandas")


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "optimizer": {
            "tolerance": 1e-5,
            "speculative": True,
            "max_functions": 3,
        },
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
            "upper_bounds": 1.0,
            "lower_bounds": -1.0,
        },
        "objective_functions": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def _handle_results(
    event: Event, reporter: ResultsDataFrame, config: EnOptConfig
) -> None:
    assert event.results is not None
    reporter.add_results(config, event.results)


def test_dataframe_results_no_results(enopt_config: Any, evaluator: Any) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(set())
    OptimizationPlanRunner(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter, config=config),
    ).run()
    assert reporter.frame.empty


def test_dataframe_results_function_results(enopt_config: Any, evaluator: Any) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(
        {
            "result_id",
            "evaluations.variables",
        },
    )
    OptimizationPlanRunner(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter, config=config),
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("evaluations.variables", idx) for idx in range(3)
    ]


def test_dataframe_results_function_results_formatted_names(
    enopt_config: Any, evaluator: Any
) -> None:
    enopt_config["variables"]["names"] = [("a", 1), ("a", 2), ("a", 3)]
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(
        {
            "result_id",
            "evaluations.variables",
        },
    )
    OptimizationPlanRunner(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter, config=config),
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("evaluations.variables", f"a:{idx}") for idx in range(1, 4)
    ]


def test_dataframe_results_gradient_results(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["variables"]["names"] = [("a", 1), ("a", 2), ("a", 3)]
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsDataFrame(
        {
            "result_id",
            "gradients.weighted_objective",
        },
        table_type="gradients",
    )
    OptimizationPlanRunner(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter, config=config),
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("gradients.weighted_objective", f"a:{idx}") for idx in range(1, 4)
    ]


def test_dataframe_results_metadata(enopt_config: Any, evaluator: Any) -> None:
    reporter = ResultsDataFrame(
        {
            "result_id",
            "evaluations.variables",
            "metadata.foo.bar",
            "metadata.not.existing",
        },
    )

    def handler(event: Event) -> None:
        assert event.results is not None
        assert event.results is not None
        for item in event.results:
            item.metadata["foo"] = {"bar": 1}
        reporter.add_results(EnOptConfig.model_validate(enopt_config), event.results)

    OptimizationPlanRunner(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, handler
    ).run()

    assert len(reporter.frame) == 3
    assert list(reporter.frame.columns.get_level_values(level=0)) == [
        ("evaluations.variables", idx) for idx in range(3)
    ] + ["metadata.foo.bar"]


def test_dataframe_results_metadata_step_id(enopt_config: Any, evaluator: Any) -> None:
    reporter = ResultsDataFrame(
        {
            "result_id",
            "evaluations.variables",
            "metadata.step_name",
        },
    )

    def handler(event: Event) -> None:
        assert event.results is not None
        reporter.add_results(EnOptConfig.model_validate(enopt_config), event.results)

    runner = OptimizationPlanRunner(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, handler
    )
    runner._plan_config["steps"][0]["name"] = "opt"  # noqa: SLF001
    runner.run()
    assert reporter.frame["metadata.step_name"].to_list() == ["opt"] * 3



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

import numpy as np  # noqa: TCH002

from ropt.enums import ResultAxisName

from ._result_field import ResultField
from ._utils import (
    _get_linear_constraint_values,
    _get_linear_constraint_violations,
    _immutable_copy,
)

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig

    from ._function_evaluations import FunctionEvaluations


@dataclass
class LinearConstraints(ResultField):
    """This class stores constraint values and violations.

    The following information is stored:

    1. Differences between the linear constraints and their right-hand-side
       values.
    2. Violations of constraints, defined as the absolute value of the
       difference if the constraint is violated, or else zero.

    Args:
        values:     Linear constraint values
        violations: Violations of the linear constraints
    """

    values: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.LINEAR_CONSTRAINT,),
        },
    )
    violations: Optional[NDArray[np.float64]] = field(
        default=None,
        metadata={
            "__axes__": (ResultAxisName.LINEAR_CONSTRAINT,),
        },
    )

    def __post_init__(self) -> None:
        """Make all array fields immutable copies.

        # noqa
        """
        self.values = _immutable_copy(self.values)
        self.violations = _immutable_copy(self.violations)

    @classmethod
    def create(
        cls, config: EnOptConfig, evaluations: FunctionEvaluations
    ) -> Optional[LinearConstraints]:
        """Add constraint information.

        This factory function creates a `LinearConstraints` object with the
        correct information on the linear constraints, based on the given
        evaluations and functions.

        Args:
            config:      The ensemble optimizer configuration object
            evaluations: An instance of the Evaluations class

        Returns:
            A newly created LinearConstraints object.
        """
        kwargs = {}
        if config.linear_constraints is not None:
            kwargs["values"] = _get_linear_constraint_values(
                config, evaluations.variables, axis=-1
            )
            kwargs["violations"] = _get_linear_constraint_violations(
                config, kwargs["values"]
            )
        return LinearConstraints(**kwargs) if kwargs else None



"""This module defines the optimization plan object."""

from __future__ import annotations

import ast
import keyword
import re
from itertools import count
from numbers import Number
from typing import TYPE_CHECKING, Any, Dict, Final, List, Optional, Union

import numpy as np
from numpy.random import default_rng

from ropt.config.enopt.constants import DEFAULT_SEED
from ropt.events import EventBroker
from ropt.exceptions import PlanError
from ropt.plugins import PluginManager

if TYPE_CHECKING:
    from ropt.config.plan import PlanConfig, StepConfig
    from ropt.evaluator import Evaluator
    from ropt.plugins.plan.base import PlanStep

_VALID_TYPES: Final = (int, float, bool)
_VALID_RESULTS: Final = (list, np.ndarray, *_VALID_TYPES)
_UNARY_OPS: Final = (ast.UAdd, ast.USub, ast.Not)
_BIN_OPS: Final = (ast.Add, ast.Sub, ast.Div, ast.FloorDiv, ast.Mult, ast.Mod, ast.Pow)
_BOOL_OPS: Final = (ast.Or, ast.And)
_CMP_OPS: Final = (ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE)


MetaDataType = Dict[str, Union[int, float, bool, str]]


class OptimizerContext:
    """Store the context in which an optimizer runs."""

    def __init__(self, evaluator: Evaluator, seed: Optional[int] = None) -> None:
        """Initialize the optimiation context.

        Args:
            evaluator:      The callable for running function evaluations
            seed:           Optional seed for the random number generator
        """
        self.evaluator = evaluator
        self.rng = default_rng(DEFAULT_SEED) if seed is None else default_rng(seed)
        self.result_id_iter = count()
        self.events = EventBroker()


class Plan:
    """The plan object."""

    def __init__(
        self,
        config: PlanConfig,
        optimizer_context: OptimizerContext,
        plugin_manager: Optional[PluginManager] = None,
        metadata: Optional[MetaDataType] = None,
    ) -> None:
        """Initialize a plan object.

        Args:
            config:            Optimizer configuration
            optimizer_context: Context in which the plan executes
            plugin_manager:    Optional plugin manager
            metadata:          Metadata to add to all results
        """
        self._plan_config = config
        self._optimizer_context = optimizer_context
        self._vars: Dict[str, Any] = {}
        self._metadata = metadata

        self._plugin_manager = (
            PluginManager() if plugin_manager is None else plugin_manager
        )
        self._context = {
            config.id: self._plugin_manager.get_plugin(
                "plan", method=config.init
            ).create(config, self)
            for config in config.context
        }
        self._steps = self.create_steps(config.steps)

    def run(self) -> bool:
        """Run the plan.

        Returns:
            Whether a user abort occurred.
        """
        return self.run_steps(self._steps)

    def create_steps(self, step_configs: List[StepConfig]) -> List[PlanStep]:
        """Create step objects from step configs.

        Args:
            step_configs: The configurations of the steps.
        """
        return [
            self._plugin_manager.get_plugin("plan", method=step_config.run).create(
                step_config, self
            )
            for step_config in step_configs
        ]

    def run_steps(self, steps: List[PlanStep]) -> bool:
        """Run the given steps.

        Args:
            steps: The steps to run

        Returns:
            Whether a user abort occurred.
        """
        return any(
            task.run() for task in steps if self._check_condition(task.step_config)
        )

    @property
    def plugin_manager(self) -> PluginManager:
        """Return the plugin manager used by the plan.

        Returns:
            The plugin manager.
        """
        return self._plugin_manager

    @property
    def optimizer_context(self) -> OptimizerContext:
        """Return the optimizer context of the plan.

        Returns:
            The optimizer context object used by the plan.
        """
        return self._optimizer_context

    @property
    def metadata(self) -> Optional[MetaDataType]:
        """Return the metadata stored by the plan.

        Returns:
            The plan metadata.
        """
        return self._metadata

    def parse_value(self, value: Any) -> Any:  # noqa: ANN401
        """Parse a value as an expression or an interpolated string.

        If the value is a string, and starts with `$`, it is assumed to be an
        expression and it is evaluated. If it does not start with `$`, any
        embedded string starting with `$` sign are evaluated and interpolated
        into the string.

        If the value is not a string, it is passed through unchanged.

        Args:
            value: The value to evaluate

        Returns:
            The result of the evaluation.
        """

        def _substitute(matched: re.Match[str]) -> str:
            value = matched.string[matched.start() : matched.end()]
            return "$" if value == "$$" else str(self.parse_value(value))

        if isinstance(value, str):
            stripped = value.strip()
            if stripped.startswith("$"):
                return self.eval(stripped)
            return re.sub(r"\${{(.*?)}}|\$\$|\$([^\W0-9][\w\.]*)", _substitute, value)
        return value

    def spawn(self, config: PlanConfig, metadata: Optional[MetaDataType]) -> Plan:
        """Spawn a child plan.

        Args:
            config:  The configuration of the new plan.
            metadata: Metadata to set in the new plan.
        """
        if metadata is not None:
            metadata = {key: self.parse_value(expr) for key, expr in metadata.items()}
        return Plan(
            config,
            optimizer_context=self._optimizer_context,
            plugin_manager=self._plugin_manager,
            metadata=metadata,
        )

    def _eval_expr(self, expr: str) -> Any:  # noqa: ANN401
        # Check for identifiers that are not preceded by $:
        for word in re.findall(r"(?<!\$)\b\w+\b", expr):
            if word.isidentifier() and not keyword.iskeyword(word):
                msg = f"Syntax error in expression: {expr}"
                raise PlanError(msg)

        # Remove $ from identifiers, before sending the string to the parser:
        stripped = expr
        for word in re.findall(r"(?<=\$)\b\w+\b", expr):
            if word.isidentifier() and not keyword.iskeyword(word):
                stripped = stripped.replace(f"${word}", word)

        # Parse the string:
        try:
            tree = ast.parse(stripped, mode="eval")
        except SyntaxError as exc:
            msg = f"Syntax error in expression: {expr}"
            raise PlanError(msg) from exc

        # Replace identifiers with their value and evaluate:
        if _is_valid(tree.body):
            replacer = _ReplaceFields(self)
            tree = ast.fix_missing_locations(replacer.visit(tree))
            try:
                result = eval(  # noqa: S307
                    compile(tree, "", mode="eval"), {"__builtins__": {}}, replacer.vars
                )
            except TypeError as exc:
                msg = f"Type error in expression: {expr}"
                raise PlanError(msg) from exc
            assert result is None or isinstance(result, _VALID_RESULTS)
            return result

        msg = f"Invalid expression: {expr}"
        raise PlanError(msg)

    def eval(self, value: Any) -> Any:  # noqa: ANN401
        """Evaluate the provided value as an expression.

        The value is evaluated as follows:

        - If the value is not a string, return unchanged
        - If the value is a string enclosed in a `${{ }}` pair the contents are
          evaluated recursively.
        - If value is a string that starts with `$` and is an identifier it is
          assumed to be a plan variable and its value is returned. The
          resulting value can have any type.
        - Otherwise the string is evaluated and the result returned. The type of
          the result is restricted to numerical values, lists and numpy array.

        Args:
            value: The expression to evaluate.

        Returns:
            The result of the expression.
        """
        if not isinstance(value, str):
            return value

        value = value.strip()

        # Recursively evaluate when enclosed in `${{ }}`:
        if value.startswith("${{") and value.endswith("}}"):
            return self.eval(value[3:-2].strip())

        # Identifiers are not evaluated, their value is returned unchanged:
        if value.startswith("$") and value[1:].isidentifier():
            return self[value[1:]]

        # Evaluate as an expression:
        return self._eval_expr(value)

    def _check_condition(self, config: StepConfig) -> bool:
        if config.if_ is not None:
            return bool(self.eval(config.if_))
        return True

    def reset_context(self, obj_id: str) -> None:
        """Reset the given context object.

        Args:
            obj_id: The ID of the object.
        """
        if obj_id in self._context:
            self._context[obj_id].reset()

    def has_context(self, obj_id: str) -> bool:
        """Check if a variable of field exists.

        Args:
            obj_id: name of the context object

        Returns:
            Whether the object exists.
        """
        return obj_id in self._context

    def update_context(self, name: str, value: Any) -> None:  # noqa: ANN401
        """Update a context object.

        Args:
            name:  The context
            value: The value to use for the update
        """
        if name not in self._context:
            msg = f"not a valid context: `{name}`"
            raise PlanError(msg)
        self._context[name].update(value)

    def __getitem__(self, name: str) -> Any:  # noqa: ANN401
        """Get the value of a variable.

        Args:
            name: the variable name

        Returns:
            The value of the variable.
        """
        if name in self._vars:
            return self._vars[name]
        msg = f"Unknown plan variable: `{name}`"
        raise PlanError(msg)

    def __setitem__(self, name: str, value: Any) -> None:  # noqa: ANN401
        """Set a variable .

        Args:
            name:  The variable
            value: The value to assign
        """
        if not name.isidentifier():
            msg = f"Not a valid variable name: `{name}`"
            raise PlanError(msg)
        self._vars[name] = value

    def __contains__(self, name: str) -> bool:
        """Check if a variable exists.

        Args:
            name: name of the variable

        Returns:
            Whether the variable exists.
        """
        return name in self._vars


def _is_valid(node: ast.AST) -> bool:  # noqa: PLR0911
    if isinstance(node, ast.Constant):
        return node.value is None or type(node.value) in _VALID_TYPES
    if isinstance(node, ast.UnaryOp):
        return isinstance(node.op, _UNARY_OPS) and _is_valid(node.operand)
    if isinstance(node, ast.BinOp):
        return (
            isinstance(node.op, _BIN_OPS)
            and _is_valid(node.left)
            and _is_valid(node.right)
        )
    if isinstance(node, ast.BoolOp):
        return (
            isinstance(node.op, _BOOL_OPS)
            and all(_is_valid(value) for value in node.values)  # noqa: PD011
        )
    if isinstance(node, ast.Compare):
        return all(isinstance(op, _CMP_OPS) for op in node.ops) and all(
            _is_valid(value) for value in node.comparators
        )
    if isinstance(node, ast.List):
        return all(_is_valid(item) for item in node.elts)
    return bool(isinstance(node, ast.Name))


class _ReplaceFields(ast.NodeTransformer):
    def __init__(self, plan: Plan) -> None:
        self._plan = plan
        self.vars: Dict[str, Any] = {}

    def visit_Name(self, node: ast.Name) -> ast.AST:  # noqa: N802
        value = self._plan[node.id]
        if value is None or isinstance(value, (Number, np.ndarray)):
            self.vars[node.id] = value
            return node
        msg = f"Error in expression: the type of `{node.id}` is not supported"
        raise PlanError(msg)



from __future__ import annotations

from dataclasses import fields
from importlib.util import find_spec
from typing import TYPE_CHECKING, Any, Dict, Final, List, Optional, Set, Tuple, Union

if TYPE_CHECKING:
    from pathlib import Path

    from ropt.results import Results

_HAVE_PANDAS: Final = find_spec("pandas") is not None
_HAVE_TABULATE: Final = find_spec("tabulate") is not None

if TYPE_CHECKING and _HAVE_PANDAS:
    import pandas as pd  # noqa: TCH002

if _HAVE_TABULATE:
    from tabulate import tabulate


def _get_select(results: Results, field_name: str, sub_fields: Set[str]) -> List[str]:
    results_field = getattr(results, field_name)
    return [
        name
        for name in {field.name for field in fields(results_field)}
        if f"{field_name}.{name}" in sub_fields
    ]


def _add_prefix(
    name: Union[Tuple[str, ...], str], prefix: str
) -> Union[Tuple[str, ...], str]:
    return (
        (f"{prefix}.{name[0]}",) + name[1:]
        if isinstance(name, tuple)
        else f"{prefix}.{name}"
    )


def _add_metadata(
    data_frame: pd.DataFrame, results: Results, sub_fields: Set[str]
) -> pd.DataFrame:
    for field in sub_fields:
        split_fields = field.split(".")
        if split_fields[0] == "metadata":
            value = _get_value(results.metadata, split_fields[1:])
            if value is not None:
                data_frame[field] = value
    return data_frame


def _get_value(data: Dict[str, Any], keys: List[str]) -> Optional[Any]:  # noqa: ANN401
    for key in keys:
        if isinstance(data, dict):
            if key not in data:
                return None
            data = data[key]
        else:
            break
    return data


def _write_table(
    table: pd.DataFrame, path: Path, min_header_len: Optional[int]
) -> None:
    if not table.empty:
        data = _align_column_names(table, min_header_len)
        table_data = {str(column): data[column] for column in data}
        path.write_text(
            tabulate(table_data, headers="keys", tablefmt="simple", showindex=False),
        )


def _extract_columns(data_frame: pd.DataFrame, mapping: Dict[str, str]) -> pd.DataFrame:
    # Columns need to be reordered to follow the ordering of the mapping keys.
    # Column names may be tuples, where the first element indicates column type
    # (e.g. `variable` or `objective`) and the rest variable or function names.
    # Hence, sets of columns exist where the first element of the column name
    # tuple is identical. Re-ordering is done according to the first element of
    # these sets of columns, ordering within the set is unchanged.
    reordered_columns = [
        name
        for key in mapping
        for name in data_frame.columns.to_numpy()
        if name == key or (isinstance(name, tuple) and name[0] == key)
    ]
    data_frame = data_frame.reindex(columns=reordered_columns)

    # Columns are renamed according to the mapping. If the column name is a
    # tuple the first element is renamed. Tuples are joined by newlines, as a
    # result column type (e.g. `variable` or `objective`) will be on the first
    # line, whereas the name of the variable or the function will be on the
    # second. Columns names that are not a tuple may contain new lines in the
    # mapped file, to split long names over multiple lines.
    renamed_columns = [
        "\n".join([mapping[name[0]]] + [str(item) for item in name[1:]])
        if isinstance(name, tuple)
        else mapping[name]
        for name in reordered_columns
    ]
    return data_frame.set_axis(renamed_columns, axis="columns")


def _align_column_names(
    data_frame: pd.DataFrame, min_header_len: Optional[int]
) -> pd.DataFrame:
    max_lines = max(len(str(column).split("\n")) for column in data_frame.columns)
    if min_header_len is not None and max_lines < min_header_len:
        max_lines = min_header_len
    return data_frame.rename(
        columns={
            column: str(column) + (max_lines - len(str(column).split("\n"))) * "\n"
            for column in data_frame.columns
        },
    )



"""Example of optimization of the Rosenbrock test function.

This example demonstrates optimization of the a modified Rosenbrock function
that exhibits uncertainty in its parameters. It shows how to write a minimal
configuration and how to run and monitor the optimization.
"""

import sys
from functools import partial
from typing import Any, Dict, List, Optional

import numpy as np
from numpy.random import default_rng
from numpy.typing import NDArray

from ropt.enums import EventType
from ropt.evaluator import EvaluatorContext, EvaluatorResult
from ropt.events import Event
from ropt.plan import OptimizationPlanRunner
from ropt.results import FunctionResults

UNCERTAINTY = 0.1


CONFIG: Dict[str, Any] = {
    "variables": {
        "initial_values": [0.5, 2.0],
    },
    "gradient": {
        "perturbation_magnitudes": 1e-7,
    },
}


def rosenbrock(
    variables: NDArray[np.float64],
    context: EvaluatorContext,
    a: NDArray[np.float64],
    b: NDArray[np.float64],
) -> EvaluatorResult:
    """Function evaluator for the rosenbrock function.

    This function returns a tuple containing the calculated objectives and
    `None`, the latter because no constraints are calculated.

    Args:
        variables: The variables to evaluate
        context:   Evaluator context
        a:         The set of parameters `a` for all realizations
        b:         The set of parameters `b` for all realizations

    Returns:
        The calculated objective, and `None`
    """
    objectives = np.zeros((variables.shape[0], 1), dtype=np.float64)
    for idx, r in enumerate(context.realizations):
        x, y = variables[idx, :]
        objectives[idx, 0] = (a[r] - x) ** 2 + b[r] * (y - x * x) ** 2
    return EvaluatorResult(objectives=objectives)


def report(event: Event) -> None:
    """Report results of an evaluation.

    Args:
        event: event data
    """
    assert event.results is not None
    for item in event.results:
        if isinstance(item, FunctionResults) and item.functions is not None:
            print(f"result: {item.result_id}")
            print(f"  variables: {item.evaluations.variables}")
            print(f"  objective: {item.functions.weighted_objective}\n")


def run_optimization(config: Dict[str, Any]) -> FunctionResults:
    """Run the optimization.

    Args:
        config: The configuration of the optimizer

    Returns:
        The optimal results.
    """
    rng = default_rng(seed=123)

    realizations = len(config["realizations"]["weights"])
    a = rng.normal(loc=1.0, scale=UNCERTAINTY, size=realizations)
    b = rng.normal(loc=100.0, scale=100 * UNCERTAINTY, size=realizations)

    optimal_result = (
        OptimizationPlanRunner(CONFIG, partial(rosenbrock, a=a, b=b))
        .add_observer(EventType.FINISHED_EVALUATION, report)
        .run()
        .results
    )
    assert optimal_result is not None
    assert optimal_result.functions is not None

    print(f"BEST RESULT: {optimal_result.result_id}")
    print(f"  variables: {optimal_result.evaluations.variables}")
    print(f"  objective: {optimal_result.functions.weighted_objective}\n")

    return optimal_result


def main(argv: Optional[List[str]] = None) -> None:
    """Run the example and check the result.

    Args:
        argv: Optional command line arguments.
    """
    if argv is not None and "--merge" in argv:
        realizations = 100
        CONFIG["realizations"] = {"weights": [1.0] * realizations}
        CONFIG["gradient"]["number_of_perturbations"] = 1
        CONFIG["gradient"]["merge_realizations"] = True
    else:
        realizations = 10
        CONFIG["realizations"] = {"weights": [1.0] * realizations}
        CONFIG["gradient"]["number_of_perturbations"] = 5

    optimal_result = run_optimization(CONFIG)
    assert optimal_result is not None
    assert optimal_result.functions is not None
    assert np.allclose(optimal_result.functions.weighted_objective, 0, atol=1e-2)
    assert np.allclose(optimal_result.evaluations.variables, 1, atol=1e-1)


if __name__ == "__main__":
    main(sys.argv)



"""The optimization plan configuration class."""

from __future__ import annotations

from collections import Counter
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, model_validator


class ContextConfig(BaseModel):
    """Configuration of a single context object.

    Context objects process information that is provided by the steps of the
    optimization plan. They usually store information in plan variables that are
    accessible to the steps and after the plan has finished. In most cases, the
    context objects store a single result in a variable with a name equal to the
    `id` of the context object, but additional variables can be defined also.

    The `init` string identifies the code that is run to initialize the context
    object. It is used by the plugin manager to load the code.

    Additional parameters needed by the context objects are configured using the
    `with_` attribute. The contents of the `with_` attribute depend on the type
    of the context object.

    Context objects are referred to by their `id`, which is mandatory.

    Note: `with` is an alias for `with_`
        When parsing dictionaries into a `ContextConfig` object the name of the
        `with_` attribute should be replaced by by `with`, i.e. without the `_`
        suffix.

    Attributes:
        id:    An identifier used to refer to the context object
        init:  Identifies the code that initializes the object
        with_: Additional parameters passed to the object
    """

    id: str
    init: str
    with_: Any = Field(default_factory=dict, alias="with")

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
    )

    @model_validator(mode="after")
    def _after_validator(self) -> ContextConfig:
        if not self.id.isidentifier():
            msg = f"Invalid ID: {self.id}"
            raise ValueError(msg)
        return self


class StepConfig(BaseModel):
    """Configuration of a single step.

    A step is a single action within an optimization plan. The `run` string
    identifies the code that executes te step. It is used by the plugin manager
    to load the code.

    A step may be named using the optional `name` field, which will only be used
    for informational purposes, such as in error messages or in generated reports.

    Additional parameters needed by the step may be configured using the `with_`
    attribute. The content of the `with_` attribute depends on the type of the
    step.

    Execution of the step can be made conditional by providing an expression via
    the `if_` attribute. The expression will be parsed and evaluated, and the
    step will only be executed if the result is `True`.

    Note: `with` and `if` aliases
        When parsing dictionaries into a `StepConfig` object the name of the
        `with_` attribute should be replaced by by `with`, and the name of the
        `if_` attribute by `if`, i.e. without the `_` suffix

    Attributes:
        name:  An optional name used to refer to the step
        run:   Identifies the code that runs the step
        with_: Additional parameters passed to the step
        if_:   Optional expression for conditional evaluation
    """

    name: Optional[str] = None
    run: str
    with_: Union[Dict[str, Any], str] = Field(default_factory=dict, alias="with")
    if_: Optional[str] = Field(default=None, alias="if")

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
    )


class PlanConfig(BaseModel):
    """Configuration for a optimization plan.

    An optimization plan consists of two sections: a context section defined
    using the `context` attribute, and a section that defines the tasks to
    perform by the `steps` attribute.

    The `context` attribute contains the configuration of the objects that
    create and maintain the environment in which the plan runs. Context objects
    are initialized before creating and running the steps.

    After initializing the context objects, the steps are configured by the
    entries given by the `steps` attribute and are initialized and executed in
    order.

    Attributes:
        context: The context objects to initialize
        steps:   The steps that are executed by the plan
    """

    context: List[ContextConfig] = []
    steps: List[StepConfig]

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
    )

    @model_validator(mode="after")
    def _after_validator(self) -> PlanConfig:
        duplicates = [
            id_
            for id_, count in Counter([item.id for item in self.context]).items()
            if count > 1
        ]
        if duplicates:
            raise ValueError("Duplicate Context ID(s): " + ", ".join(duplicates))
        return self



"""Configuration class for variables."""

from __future__ import annotations

from itertools import zip_longest
from typing import Optional, Tuple, Union

import numpy as np
from pydantic import Field, model_validator

from ropt.config.utils import (
    Array1D,
    ArrayEnum,
    ArrayIndices,
    UniqueNames,
    broadcast_1d_array,
    broadcast_arrays,
    check_enum_values,
    immutable_array,
)
from ropt.enums import VariableType

from ._enopt_base_model import EnOptBaseModel


class VariablesConfig(EnOptBaseModel):
    r"""The configuration class for variables.

    This configuration class, configured by the `variables` field in an
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object defines essential
    aspects of the variables: the initial values and the bounds. These are given
    by the `initial_values`, `lower_bounds`, and `upper_bounds` fields, which
    are [`numpy`](https://numpy.org) arrays. Initial values must be provided,
    while bounds are set to $-\infty$ and $+\infty$ by default. The
    `lower_bounds` and `upper_bounds` fields may contain `numpy.nan` values,
    indicating that corresponding variables have no lower or upper bounds,
    respectively. These values are converted to `numpy.inf` values with an
    appropriate sign.

    The `names` field is optional since variable names are not strictly needed
    by the optimizer. However, if the `names` field is given, it determines the
    number of variables, and the `initial_values`, `lower_bounds`, and
    `upper_bounds` fields are broadcasted accordingly. If `names` is not
    provided, these fields are broadcasted to each other, and their final size
    determines the number of variables.

    Info: Variable Names
        `ropt` does not use the names itself, and names can be of arbitrary
        type, as long as they are unique. However, some optimizers or external
        code might need a string representation of each name, which can be
        obtained using the
        [`get_formatted_names`][ropt.config.enopt.VariablesConfig.get_formatted_names]
        method. The `delimiters` attribute is used by this method to convert the
        special case of names consisting of tuples of strings.

    The optional `types` field can be used to assign types to each variable,
    according to the [`VariableType`][ropt.enums.VariableType] enumeration. The
    values can be used to configure the optimizer accordingly. If not provided,
    all variables are assumed to be continuous and of real data type
    (corresponding to [`VariableType.REAL`][ropt.enums.VariableType.REAL])

    The `offsets` and `scales` fields are optional: if given, they are
    broadcasted to the number of variables and used for scaling. The elements
    $x_i$ of `initial_values`, `lower_bounds`, and `upper_bounds` fields are
    rescaled by the elements $o_i$ and $s_i$ of `offsets` and `scales`: $(x_i -
    o_i) / s_i$.

    Info: Transformation of Linear Constraints
        Any linear constraints defined in the `EnOptConfig` object via its
        `linear_constraints` field will also be transformed using any offsets
        and scales passed via the `VariablesConfig` object. See:
        [`LinearConstraintsConfig`][ropt.config.enopt.LinearConstraintsConfig].

    The optional `indices` field contains the indices of the variables
    considered to be free to change. During optimization, only these variables
    should change while others remain fixed.

    Attributes:
        names:          Optional names of the variables
        types:          The type of the variables (optional).
        initial_values: The initial values of the variables
        lower_bounds:   Lower bound of the variables (default: $-\infty$)
        upper_bounds:   Upper bound of the variables (default: $+\infty$)
        offsets:        Optional offsets, used for scaling the variables
        scales:         Optional scales, used for scaling the variables
        indices:        Optional indices of variables to optimize.
        delimiters:     Delimiters used to construct names from tuples
    """

    names: Optional[UniqueNames] = None
    types: Optional[ArrayEnum] = None
    initial_values: Array1D = np.array(0.0)
    lower_bounds: Array1D = np.array(-np.inf)
    upper_bounds: Array1D = np.array(np.inf)
    offsets: Optional[Array1D] = None
    scales: Optional[Array1D] = None
    indices: Optional[ArrayIndices] = None
    delimiters: str = Field(":", min_length=0)

    @model_validator(mode="after")
    def _broadcast_and_scale(self) -> "VariablesConfig":
        if self.names is not None:
            size = len(self.names)
            self.initial_values = broadcast_1d_array(
                self.initial_values, "initial_values", size
            )
            self.lower_bounds = broadcast_1d_array(
                self.lower_bounds, "lower_bounds", size
            )
            self.upper_bounds = broadcast_1d_array(
                self.upper_bounds, "upper_bounds", size
            )
        else:
            (
                self.initial_values,
                self.lower_bounds,
                self.upper_bounds,
            ) = broadcast_arrays(
                self.initial_values, self.lower_bounds, self.upper_bounds
            )

        if self.offsets is not None:
            self.offsets = broadcast_1d_array(
                self.offsets, "offsets", self.initial_values.size
            )
            self.initial_values = immutable_array(self.initial_values - self.offsets)
            self.lower_bounds = immutable_array(self.lower_bounds - self.offsets)
            self.upper_bounds = immutable_array(self.upper_bounds - self.offsets)
        if self.scales is not None:
            self.scales = broadcast_1d_array(
                self.scales, "scales", self.initial_values.size
            )
            self.initial_values = immutable_array(self.initial_values / self.scales)
            self.lower_bounds = immutable_array(self.lower_bounds / self.scales)
            self.upper_bounds = immutable_array(self.upper_bounds / self.scales)

        if self.types is not None:
            check_enum_values(self.types, VariableType)
            self.types = broadcast_1d_array(
                self.types, "types", self.initial_values.size
            )

        return self

    def get_formatted_names(self) -> Optional[Tuple[str, ...]]:
        """Return string representations of the variable names.

        This method converts the variable names to a tuple of strings. Each name
        is converted using its string representation unless the name is a tuple.
        In that case, the tuple items are converted to strings and joined using
        the delimiters taken from the `delimiter` field in the `Variables`
        object. This field is a string that may consist of multiple delimiters
        used in turn. If it contains fewer items than needed, the last one is
        used for the missing ones. By default, the `:` character is used as the
        delimiter.

        Returns:
            A tuple of formatted variable names.
        """
        if self.names is None:
            return None

        def _format_name(name: Union[str, Tuple[str, ...]], delimiters: str) -> str:
            if isinstance(name, tuple):
                if not delimiters:
                    return "".join(str(item) for item in name)
                truncated_delimiters = delimiters[: len(name) - 1]
                return name[0] + "".join(
                    str(item)
                    for item_tuple in zip_longest(
                        truncated_delimiters, name[1:], fillvalue=delimiters[-1]
                    )
                    for item in item_tuple
                )
            return str(name)

        return tuple(_format_name(name, self.delimiters) for name in self.names)



"""This module implements the default optimizer step."""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import numpy as np
from pydantic import BaseModel, ConfigDict

from ropt.config.enopt import EnOptConfig
from ropt.config.plan import PlanConfig  # noqa: TCH001
from ropt.config.utils import Array1D  # noqa: TCH001
from ropt.ensemble_evaluator import EnsembleEvaluator
from ropt.enums import EventType, OptimizerExitCode
from ropt.exceptions import PlanError
from ropt.plan import ContextUpdateResults, EnsembleOptimizer, MetaDataType, Plan
from ropt.plugins.plan.base import PlanStep
from ropt.results import FunctionResults

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ropt.config.plan import StepConfig
    from ropt.results import Results


class DefaultNestedPlan(BaseModel):
    """Parameters used by the nested optimizer step.

    Attributes:
        plan:        Optional nested plan to run during optimization
        initial_var: Name of the variable providing initial_values to the nested plan
        results_var: The name of the variable in the plan containing the result
    """

    plan: PlanConfig
    initial_var: str
    results_var: str


class DefaultOptimizerStepWith(BaseModel):
    """Parameters used by the default optimizer step.

    Optionally the initial variables to be used can be set from a context object.

    Attributes:
        config:         ID of the context object that contains the optimizer configuration
        update:         List of the objects that are notified of new results
        initial_values: The initial values for the optimizer
        metadata:       Metadata to set in the results
        exit_code_var:  Name of the variable to store the exit code
        nested_plan:    Optional nested plan configuration
    """

    config: str
    update: List[str] = []
    initial_values: Optional[Union[str, Array1D]] = None
    metadata: MetaDataType = {}
    exit_code_var: Optional[str] = None
    nested_plan: Optional[DefaultNestedPlan] = None

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
        arbitrary_types_allowed=True,
    )


class DefaultOptimizerStep(PlanStep):
    """The default optimizer step."""

    def __init__(self, config: StepConfig, plan: Plan) -> None:
        """Initialize a default optimizer step.

        Args:
            config: The configuration of the step
            plan:   The plan that runs this step
        """
        super().__init__(config, plan)

        self._with = (
            DefaultOptimizerStepWith.model_validate({"config": config.with_})
            if isinstance(config.with_, str)
            else DefaultOptimizerStepWith.model_validate(config.with_)
        )
        self._enopt_config: EnOptConfig

    def run(self) -> bool:
        """Run the optimizer step.

        Returns:
            Whether a user abort occurred.
        """
        config = self.plan.parse_value(self._with.config)
        if not isinstance(config, (dict, EnOptConfig)):
            msg = "No valid EnOpt configuration provided"
            raise PlanError(msg, step_name=self.step_config.name)
        self._enopt_config = EnOptConfig.model_validate(config)

        self.plan.optimizer_context.events.emit(
            EventType.START_OPTIMIZER_STEP,
            config=self._enopt_config,
            step_name=self.step_config.name,
        )

        assert self.plan.optimizer_context.rng is not None
        ensemble_evaluator = EnsembleEvaluator(
            self._enopt_config,
            self.plan.optimizer_context.evaluator,
            self.plan.optimizer_context.result_id_iter,
            self.plan.optimizer_context.rng,
            self.plan.plugin_manager,
        )

        variables = self._get_variables()
        exit_code = EnsembleOptimizer(
            enopt_config=self._enopt_config,
            ensemble_evaluator=ensemble_evaluator,
            plugin_manager=self.plan.plugin_manager,
            nested_optimizer=self._run_nested_plan,
            signal_evaluation=self._signal_evaluation,
        ).start(variables)

        if self._with.exit_code_var is not None:
            self.plan[self._with.exit_code_var] = exit_code

        self.plan.optimizer_context.events.emit(
            EventType.FINISHED_OPTIMIZER_STEP,
            config=self._enopt_config,
            exit_code=exit_code,
            step_name=self.step_config.name,
        )

        return exit_code == OptimizerExitCode.USER_ABORT

    def _signal_evaluation(self, results: Optional[Tuple[Results, ...]] = None) -> None:
        """Called before and after the optimizer finishes an evaluation.

        Before the evaluation starts, this method is called with the `results`
        argument set to `None`. When an evaluation is has finished, it is called
        with `results` set to the results of the evaluation.

        Args:
            results: The results produced by the evaluation.
        """
        if results is None:
            self.plan.optimizer_context.events.emit(
                EventType.START_EVALUATION,
                config=self._enopt_config,
                step_name=self.step_config.name,
            )
        else:
            metadata = self._get_metadata(add_step_name=True)
            for item in results:
                item.metadata = metadata

            for obj_id in self._with.update:
                self.plan.update_context(
                    obj_id,
                    ContextUpdateResults(
                        step_name=self.step_config.name,
                        results=results,
                    ),
                )
            self.plan.optimizer_context.events.emit(
                EventType.FINISHED_EVALUATION,
                config=self._enopt_config,
                results=results,
                step_name=self.step_config.name,
            )

    def _run_nested_plan(
        self, variables: NDArray[np.float64]
    ) -> Tuple[Optional[FunctionResults], bool]:
        """Run a  nested plan.

        Args:
            variables: variables to set in the nested plan.

        Returns:
            The variables generated by the nested plan.
        """
        if self._with.nested_plan is None:
            return None, False
        metadata = self._get_metadata(add_step_name=False)
        plan = self.plan.spawn(self._with.nested_plan.plan, metadata)
        plan[self._with.nested_plan.initial_var] = variables
        aborted = plan.run()
        return plan[self._with.nested_plan.results_var], aborted

    def _get_variables(self) -> NDArray[np.float64]:
        if self._with.initial_values is not None:
            parsed_variables = self.plan.parse_value(self._with.initial_values)
            if isinstance(parsed_variables, FunctionResults):
                return parsed_variables.evaluations.variables
            if isinstance(parsed_variables, np.ndarray):
                return parsed_variables
            if parsed_variables is not None:
                msg = f"`{self._with.initial_values} does not contain variables."
                raise PlanError(msg, step_name=self.step_config.name)
        return self._enopt_config.variables.initial_values

    def _get_metadata(self, *, add_step_name: bool) -> MetaDataType:
        metadata = self.plan.metadata
        if metadata is None:
            metadata = {}
        if add_step_name and self.step_config.name is not None:
            metadata["step_name"] = self.step_config.name
        for key, expr in self._with.metadata.items():
            metadata[key] = self.plan.parse_value(expr)
        return metadata



"""Ensemble optimizer class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Protocol, Tuple

import numpy as np

from ropt.enums import ConstraintType, OptimizerExitCode
from ropt.exceptions import OptimizationAborted
from ropt.results import (
    BoundConstraints,
    FunctionResults,
    GradientResults,
    LinearConstraints,
    NonlinearConstraints,
)

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ropt.config.enopt import EnOptConfig
    from ropt.ensemble_evaluator import EnsembleEvaluator
    from ropt.plugins import PluginManager
    from ropt.results import Functions, Gradients, Results


class SignalEvaluationCallback(Protocol):
    """Protocol for the callback that is used to signal that an evaluation is occurring."""

    def __call__(self, results: Optional[Tuple[Results, ...]] = None, /) -> None:
        """Callback protocal.

        When a function with this signature is provided to an ensemble
        optimizer, it is called before and after the optimizer finishes an
        evaluation.Before the evaluation starts, this method is called with the
        `results` argument set to `None`. When an evaluation is has finished, it
        is called with `results` set to the results of the evaluation.

        Args:
            results: The results produced by the evaluation.
        """


class NestedOptimizerCallback(Protocol):
    """Run a nested optimization."""

    def __call__(
        self, variables: NDArray[np.float64], /
    ) -> Tuple[Optional[FunctionResults], bool]:
        """Callback protocol.

        When a function with this signature is provided to an ensemble
        optimizer, it is called at each function evaluation to run a nested
        optimization. It accepts the current variables as its arguments and
        returns a tuple consisting of the result (or `None` if no result is
        available), and a boolean value that indicates if the optimization was
        aborted by a user signal.

        Args:
            variables: The variables to start the nested optimization with

        Returns:
            The nested optimization results and whether the user aborted the optimization.
        """


class EnsembleOptimizer:
    """Optimizer class for running ensemble optimizations."""

    def __init__(
        self,
        *,
        enopt_config: EnOptConfig,
        ensemble_evaluator: EnsembleEvaluator,
        plugin_manager: PluginManager,
        signal_evaluation: Optional[SignalEvaluationCallback] = None,
        nested_optimizer: Optional[NestedOptimizerCallback] = None,
    ) -> None:
        """Initialize the ensemble optimizer class.

        This class requires at least three argument thats together define a single optimization run:

        1. An [`EnOptConfig`][ropt.config.enopt.EnOptConfig] that contains all
           configuration settings for the optimization.
        2. An [`EnsembleEvaluator`][ropt.evaluator.EnsembleEvaluator] object
           that is responsible for evaluating functions and gradients.
        3. A [`PluginManager`][ropt.plugins.PluginManager] object that provides
           access to optimizer plugins.

        In addition, two optional callbacks can be provided:

        1. A [`SignalEvaluationCallback`][ropt.optimizer.SignalEvaluationCallback]
           callback that is called before and after each function evaluation.
        2. A [`NestedOptimizerCallback`][ropt.optimizer.NestedOptimizerCallback]
           callback that is called at each function evaluation to run a nested
           optimization.

        Args:
            enopt_config:       The ensemble optimization configuration
            ensemble_evaluator: The evaluator object
            plugin_manager:     Plugin manager
            signal_evaluation:  Optional callback to signal evaluations
            nested_optimizer:   Optional nested optimization call
        """
        self._enopt_config = enopt_config
        self._function_evaluator = ensemble_evaluator
        self._plugin_manager = plugin_manager
        self._signal_evaluation = signal_evaluation
        self._nested_optimizer = nested_optimizer

        # This stores the values of the fixed variable
        self._fixed_variables: NDArray[np.float64]

        # For implementing max_functions:
        self._completed_functions = 0

        # Whether NaN values are allowed:
        self._allow_nan = False

    def start(self, variables: NDArray[np.float64]) -> OptimizerExitCode:
        """Start the optimization.

        Args:
            variables: The initial variables for the optimization.

        Returns:
            _description_
        """
        self._fixed_variables = variables.copy()

        optimizer = self._plugin_manager.get_plugin(
            "optimizer", method=self._enopt_config.optimizer.method
        ).create(self._enopt_config, self._optimizer_callback)
        self._allow_nan = optimizer.allow_nan

        exit_code = OptimizerExitCode.OPTIMIZER_STEP_FINISHED
        try:
            optimizer.start(variables)
        except OptimizationAborted as exc:
            exit_code = exc.exit_code
        return exit_code

    def _optimizer_callback(
        self,
        variables: NDArray[np.float64],
        *,
        return_functions: bool,
        return_gradients: bool,
    ) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        assert return_functions or return_gradients

        self._check_stopping_criteria()

        variables = self._get_completed_variables(variables)

        # Run any nested steps, when this improves the objective, this may
        # change the fixed variables and the current optimal result:
        if self._nested_optimizer is not None:
            nested_results, aborted = self._nested_optimizer(variables)
            if nested_results is not None:
                if aborted:
                    raise OptimizationAborted(exit_code=OptimizerExitCode.USER_ABORT)
                variables = nested_results.evaluations.variables.copy()
                self._fixed_variables = variables.copy()

        results = self._run_evaluations(
            variables,
            compute_functions=return_functions,
            compute_gradients=return_gradients,
        )

        # Functions and gradients might need to be scaled:
        scales, offsets = self._get_scale_parameters()

        functions = np.array([])
        if return_functions:
            # Functions might be parallelized hence we need potentially to
            # process a list of function results:
            functions_list = [
                self._scale_constraints(
                    self._functions_from_results(item.functions), scales, offsets
                )
                for item in results
                if isinstance(item, FunctionResults)
            ]
            self._completed_functions += len(functions_list)
            functions = (
                np.vstack(functions_list) if variables.ndim > 1 else functions_list[0]
            )

        gradients = np.array([])
        if return_gradients:
            # Gradients cannot be parallelized, there is at most one gradient:
            gradients = self._scale_constraints(
                self._gradients_from_results(
                    next(
                        item.gradients
                        for item in results
                        if isinstance(item, GradientResults)
                    ),
                    self._enopt_config.variables.indices,
                ),
                scales,
            )

        return functions, gradients

    def _get_completed_variables(
        self, variables: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        indices = self._enopt_config.variables.indices
        if indices is not None:
            if variables.ndim > 1:
                tmp_variables = np.repeat(
                    variables[np.newaxis, :], variables.shape[0], axis=0
                )
                tmp_variables[:, indices] = variables
            else:
                tmp_variables = self._fixed_variables.copy()
                tmp_variables[indices] = variables
            return tmp_variables
        return variables.copy()

    def _check_stopping_criteria(self) -> None:
        max_functions = self._enopt_config.optimizer.max_functions
        if max_functions is not None and self._completed_functions >= max_functions:
            raise OptimizationAborted(exit_code=OptimizerExitCode.MAX_FUNCTIONS_REACHED)

    def _run_evaluations(
        self,
        variables: NDArray[np.float64],
        *,
        compute_functions: bool = False,
        compute_gradients: bool = False,
    ) -> Tuple[Results, ...]:
        assert compute_functions or compute_gradients
        if self._signal_evaluation:
            self._signal_evaluation()
        results = self._function_evaluator.calculate(
            variables,
            compute_functions=compute_functions,
            compute_gradients=compute_gradients,
        )
        results = self._augment_results(results)
        if self._signal_evaluation:
            self._signal_evaluation(results)

        # If the configuration allows for zero successful realizations, there
        # will always be results. However, they may all be equal to `np.nan`. If
        # the optimizer does not set the allow_nan flag, it cannot handle such a
        # case, and we need to check for it:
        assert self._enopt_config.realizations.realization_min_success is not None
        check_failures = (
            self._enopt_config.realizations.realization_min_success < 1
            and not self._allow_nan
        )
        for result in results:
            assert isinstance(result, (FunctionResults, GradientResults))
            if (
                (isinstance(result, FunctionResults) and result.functions is None)
                or (isinstance(result, GradientResults) and result.gradients is None)
                or (check_failures and np.all(result.realizations.failed_realizations))
            ):
                raise OptimizationAborted(
                    exit_code=OptimizerExitCode.TOO_FEW_REALIZATIONS
                )

        return results

    def _augment_results(self, results: Tuple[Results, ...]) -> Tuple[Results, ...]:
        for result in results:
            if isinstance(result, FunctionResults):
                result.bound_constraints = BoundConstraints.create(
                    self._enopt_config, result.evaluations
                )
                result.linear_constraints = LinearConstraints.create(
                    self._enopt_config, result.evaluations
                )
                result.nonlinear_constraints = NonlinearConstraints.create(
                    self._enopt_config,
                    result.functions,
                    self._function_evaluator.constraint_auto_scales,
                )
        return results

    @staticmethod
    def _functions_from_results(functions: Optional[Functions]) -> NDArray[np.float64]:
        assert functions is not None
        return (
            np.array(functions.weighted_objective, ndmin=1)
            if functions.constraints is None
            else np.append(functions.weighted_objective, functions.constraints)
        )

    @staticmethod
    def _gradients_from_results(
        gradients: Optional[Gradients], variable_indices: Optional[NDArray[np.intc]]
    ) -> NDArray[np.float64]:
        assert gradients is not None
        weighted_objective_gradient = (
            gradients.weighted_objective.copy()
            if variable_indices is None
            else gradients.weighted_objective[variable_indices]
        )
        constraint_gradients = (
            None
            if gradients.constraints is None
            else (
                gradients.constraints.copy()
                if variable_indices is None
                else gradients.constraints[:, variable_indices]
            )
        )
        return (
            np.expand_dims(weighted_objective_gradient, axis=0)
            if constraint_gradients is None
            else np.vstack((weighted_objective_gradient, constraint_gradients))
        )

    def _get_constraint_scales(self, config: EnOptConfig) -> NDArray[np.float64]:
        assert config.nonlinear_constraints is not None
        scales = self._function_evaluator.constraint_auto_scales
        if scales is None:
            return config.nonlinear_constraints.scales
        return config.nonlinear_constraints.scales * scales

    def _get_scale_parameters(
        self,
    ) -> Tuple[Optional[NDArray[np.float64]], Optional[NDArray[np.float64]]]:
        if self._enopt_config.nonlinear_constraints is not None:
            offsets = self._enopt_config.nonlinear_constraints.rhs_values
            scales = self._get_constraint_scales(self._enopt_config)
            scales = np.where(
                self._enopt_config.nonlinear_constraints.types == ConstraintType.GE,
                -scales,
                scales,
            )
            return scales, offsets
        return None, None

    def _scale_constraints(
        self,
        functions: NDArray[np.float64],
        scales: Optional[NDArray[np.float64]],
        offsets: Optional[NDArray[np.float64]] = None,
    ) -> NDArray[np.float64]:
        if functions.size > 1:
            if offsets is not None:
                functions[1:, ...] = functions[1:, ...] - offsets
            if scales is not None:
                if functions.ndim > 1:
                    scales = scales[:, np.newaxis]
                functions[1:, ...] = functions[1:, ...] / scales
        return functions



"""Data classes for storing intermediate optimization results.

During the optimization process, functions and gradients are calculated and need
to be reported. To streamline this process, after new results become available,
these are passed to callbacks as a tuple of [`Results`][ropt.results.Results]
objects. These may be instances of the derived
[`FunctionResults`][ropt.results.FunctionResults] or
[`GradientResults`][ropt.results.GradientResults] classes, containing results
for function and gradient evaluations, respectively.

These classes are nested `dataclasses` with some added features:

- Variables may have scaling and offset factors associated with them. The
  optimization will generally deal with scaled values. When the results object
  is created, and variable scaling is configured, the values are scaled back and
  added to the results object.
- Objective and constraint function values may also have a scaling factor
  associated with them. These are generally handled in unscaled form, and when
  the report object is created, scaled values are added.
- Methods are available to export the results to
  [`pandas`](https://pandas.pydata.org/) data frames or
  [`xarray`](https://xarray.dev/) data sets.
- Methods are added to write and read netCDF version 4 files using the
  [`netcdf4`](https://unidata.github.io/netcdf4-python/) Python package.

Much of the data stored in the result objects is of a multi-dimensional nature.
For instance, the `objectives` field, which is part of the nested
[`evaluations`][ropt.results.FunctionEvaluations] object in the
[`FunctionResults`][ropt.results.FunctionResults] object, is a two-dimensional
`numpy` array. In this array, the columns correspond to the objectives, and the
rows correspond to the realization number.

To facilitate exporting and reporting the results, the identity of the axes in
such multi-dimensional arrays is stored in metadata associated with the
corresponding field. These fields derive from the
[`ResultField`][ropt.results.ResultField] class, which has a
[`get_axis_names`][ropt.results.ResultField.get_axis_names] class method to
retrieve the names. For the `objectives` example above, this retrieves the axis
names:

```py
>>> from ropt.results import FunctionEvaluations
>>> FunctionEvaluations.get_axis_names("objectives")
(<ResultAxisName.REALIZATION: 'realization'>, <ResultAxisName.OBJECTIVE: 'objective'>)
```

Using this metadata, the exporting or reporting code can refer to the optimizer
configuration to associate realization and objective names with any entry in the
result matrix. For instance, the [`pandas`](https://pandas.pydata.org/)
exporting code will use this to construct a multi-index for the generated data
frame and optionally unstack such multi-dimensional data into multiple columns.
"""

from ._bound_constraints import BoundConstraints
from ._function_evaluations import FunctionEvaluations
from ._function_results import FunctionResults
from ._functions import Functions
from ._gradient_evaluations import GradientEvaluations
from ._gradient_results import GradientResults
from ._gradients import Gradients
from ._linear_constraints import LinearConstraints
from ._maximize import convert_to_maximize
from ._nonlinear_constraints import NonlinearConstraints
from ._realizations import Realizations
from ._result_field import ResultField
from ._results import Results

__all__ = [
    "BoundConstraints",
    "FunctionEvaluations",
    "FunctionResults",
    "Functions",
    "GradientEvaluations",
    "GradientResults",
    "Gradients",
    "LinearConstraints",
    "NonlinearConstraints",
    "Realizations",
    "Results",
    "ResultField",
    "convert_to_maximize",
]



"""Generate a results report in a `pandas` data frame."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Literal, Set, Tuple

from ropt.enums import ResultAxisName
from ropt.results import FunctionResults, GradientResults

from ._utils import _HAVE_PANDAS, _add_metadata, _add_prefix, _get_select

if TYPE_CHECKING:
    from ropt.config.enopt import EnOptConfig
    from ropt.results import Results

if _HAVE_PANDAS:
    import pandas as pd


class ResultsDataFrame:
    """Generate a results report in a `pandas` DataFrame.

    The class is intended to be used for gathering results from
    [`Results`][ropt.results.Results] objects and storing them as a
    [`pandas`](https://pandas.pydata.org/) DataFrame.

    New results can be added to the stored DataFrame as they become available
    using the `add_results` method. The content of the DataFrame is taken from the
    fields of the [`Results`][ropt.results.Results] objects passed during each call
    to `add_results`. The updated table can be retrieved at any time via the
    [`frame`][ropt.report.ResultsDataFrame.frame] property.
    """

    def __init__(
        self,
        fields: Set[str],
        *,
        table_type: Literal["functions", "gradients"] = "functions",
    ) -> None:
        """Initialize a ResultsDataFrame object.

        The set of fields used is determined by the `results_fields` argument
        passed when initializing the object. These can be any of the fields
        defined in a [`FunctionResults`][ropt.results.FunctionResults] or a
        [`GradientResults`][ropt.results.GradientResults]. Most of the fields in
        these objects are nested, and the fields to export must be specified
        specifically in the form `field.subfield`. For instance, to specify the
        `variables` field of the `evaluations` field from a function result,
        the specification would be `evaluations.variables`.

        Note that many fields may, in fact, generate multiple columns in the
        resulting data frame. For instance, when specifying
        `evaluations.variables`, a column will be generated for each variable.
        If available, names specified in the optimizer configuration, such as
        variable names, will be used as column labels. Because the exported fields
        may be multi-dimensional with names defined along each axis, for instance,
        realizations and objectives, which both can be named, the final name may
        consist of a tuple of names.

        The `table_type` argument is used to determine which type of results
        should be reported: either function evaluation results (`functions`) or
        gradient results (`gradients`).

        Args:
            fields:     The fields of the results to store
            table_type: The type of the table
        """
        if not _HAVE_PANDAS:
            msg = "ResultsDataFrame requires the `pandas` modules"
            raise NotImplementedError(msg)

        self._fields = fields
        self._table_type = table_type
        self._frame = pd.DataFrame()

    def add_results(self, config: EnOptConfig, results: Tuple[Results, ...]) -> bool:
        """Add results to the table.

        This method can be called directly from any observers connected to
        events that produce results.

        Args:
            config:  The configuration of the optimizer generating the results
            results: The results to add

        Returns:
            True if a result was added, else False
        """
        added = False
        for item in results:
            if (
                self._table_type == "functions"
                and isinstance(item, FunctionResults)
                and item.functions is not None
            ):
                frame = _get_function_results(config, item, self._fields)
            elif (
                self._table_type == "gradients"
                and isinstance(item, GradientResults)
                and item.gradients is not None
            ):
                frame = _add_gradient_results(config, item, self._fields)
            else:
                continue
            if not frame.empty:
                frame = _add_metadata(frame, item, self._fields)
                self._frame = pd.concat([self._frame, frame])
                added = True
        return added

    @property
    def frame(self) -> pd.DataFrame:
        """Return the function results generated so far.

        Returns:
            A pandas data frame with the results.
        """
        return self._frame


def _get_function_results(
    config: EnOptConfig, results: Results, sub_fields: Set[str]
) -> pd.DataFrame:
    if (
        not sub_fields
        or not isinstance(results, FunctionResults)
        or results.functions is None
    ):
        return pd.DataFrame()

    functions = results.to_dataframe(
        config,
        "functions",
        select=_get_select(results, "functions", sub_fields),
        unstack=[ResultAxisName.OBJECTIVE, ResultAxisName.NONLINEAR_CONSTRAINT],
    ).rename(columns=partial(_add_prefix, prefix="functions"))

    bound_constraints = (
        pd.DataFrame()
        if results.bound_constraints is None
        else results.to_dataframe(
            config,
            "bound_constraints",
            select=_get_select(results, "bound_constraints", sub_fields),
            unstack=[ResultAxisName.VARIABLE],
        ).rename(columns=partial(_add_prefix, prefix="bound_constraints"))
    )

    linear_constraints = (
        pd.DataFrame()
        if results.linear_constraints is None
        else results.to_dataframe(
            config,
            "linear_constraints",
            select=_get_select(results, "linear_constraints", sub_fields),
            unstack=[ResultAxisName.LINEAR_CONSTRAINT],
        ).rename(columns=partial(_add_prefix, prefix="linear_constraints"))
    )

    nonlinear_constraints = (
        pd.DataFrame()
        if results.nonlinear_constraints is None
        else results.to_dataframe(
            config,
            "nonlinear_constraints",
            select=_get_select(results, "nonlinear_constraints", sub_fields),
            unstack=[ResultAxisName.NONLINEAR_CONSTRAINT],
        ).rename(columns=partial(_add_prefix, prefix="nonlinear_constraints"))
    )

    evaluations = results.to_dataframe(
        config,
        "evaluations",
        select=_get_select(results, "evaluations", sub_fields),
        unstack=[
            ResultAxisName.VARIABLE,
            ResultAxisName.OBJECTIVE,
            ResultAxisName.NONLINEAR_CONSTRAINT,
        ],
    ).rename(columns=partial(_add_prefix, prefix="evaluations"))

    return _join_frames(
        functions,
        bound_constraints,
        linear_constraints,
        nonlinear_constraints,
        evaluations,
    )


def _add_gradient_results(
    config: EnOptConfig, results: Results, sub_fields: Set[str]
) -> pd.DataFrame:
    if (
        not sub_fields
        or not isinstance(results, GradientResults)
        or results.gradients is None
    ):
        return pd.DataFrame()

    gradients = results.to_dataframe(
        config,
        "gradients",
        select=_get_select(results, "gradients", sub_fields),
        unstack=[
            ResultAxisName.OBJECTIVE,
            ResultAxisName.NONLINEAR_CONSTRAINT,
            ResultAxisName.VARIABLE,
        ],
    ).rename(columns=partial(_add_prefix, prefix="gradients"))

    evaluations = results.to_dataframe(
        config,
        "evaluations",
        select=_get_select(results, "evaluations", sub_fields),
        unstack=[
            ResultAxisName.VARIABLE,
            ResultAxisName.OBJECTIVE,
            ResultAxisName.NONLINEAR_CONSTRAINT,
        ],
    ).rename(columns=partial(_add_prefix, prefix="evaluations"))

    return _join_frames(gradients, evaluations)


def _join_frames(*args: pd.DataFrame) -> pd.DataFrame:
    frames = [frame for frame in args if not frame.empty]
    return (
        frames[0].join(list(frames[1:]), how="outer") if len(frames) > 1 else frames[0]
    )



import importlib
import os
import shutil
from pathlib import Path
from typing import Any, Optional

import pytest


def _load_from_file(name: str, sub_path: Optional[str] = None) -> Any:
    path = Path(__file__).parent.parent / "examples"
    if sub_path is not None:
        path = path / sub_path
    path = path / f"{name}.py"
    spec = importlib.util.spec_from_file_location(name, path)
    assert spec is not None
    module = importlib.util.module_from_spec(spec)
    assert spec.loader is not None
    spec.loader.exec_module(module)
    return module


def test_rosenbrock_deterministic(tmp_path: Path) -> None:
    os.chdir(tmp_path)
    module = _load_from_file("rosenbrock_deterministic")
    module.main()


def test_rosenbrock_ensemble(tmp_path: Path) -> None:
    os.chdir(tmp_path)
    module = _load_from_file("rosenbrock_ensemble")
    module.main()


def test_rosenbrock_ensemble_merged(tmp_path: Path) -> None:
    os.chdir(tmp_path)
    module = _load_from_file("rosenbrock_ensemble")
    module.main(["--merge"])


def test_differential_evolution_linear(tmp_path: Path) -> None:
    os.chdir(tmp_path)
    module = _load_from_file("de_linear")
    module.main()


def test_differential_evolution_nonlinear(tmp_path: Path) -> None:
    os.chdir(tmp_path)
    module = _load_from_file("de_nonlinear")
    module.main()


@pytest.mark.slow()
@pytest.mark.filterwarnings("ignore::pytest.PytestUnraisableExceptionWarning")
def test_script_optimizer(tmp_path: Path) -> None:
    os.chdir(tmp_path)
    module = _load_from_file("run", sub_path="script_optimizer")
    shutil.copyfile(
        Path(__file__).parent.parent
        / "examples"
        / "script_optimizer"
        / "rosenbrock.py",
        tmp_path / "rosenbrock.py",
    )
    module.main()



r"""Function Evaluations.

An optimizer will repeatedly request function and gradient evaluations. In each
iteration, it may require values for multiple variable vectors, and the
gradients may be calculated from various perturbed variable vectors. To address
this, the optimizer requires a function evaluator callback upon initialization,
conforming to the [Evaluator][ropt.evaluator.Evaluator] signature.

The callback accepts a matrix containing multiple variable vectors to evaluate
together with a [`EvaluatorContext`][ropt.evaluator.EvaluatorContext] object
providing information usable by the evaluator. It returns an
[`EvaluatorResult`][ropt.evaluator.EvaluatorResult] object with the calculated
values of the objective and constraint functions for all variable vectors and
realizations, and optionally some additional metadata.

Tip: Reusing Objective and Non-linear Constraint Values
    When defining multiple objectives, there might be an intention to reuse the
    same objective value multiple times. For example, a total objective could
    consist of the mean of the objectives for each realization, plus the
    standard deviation of the same values. This can be implemented by defining
    two objectives: the first calculated as the mean of the realizations, and
    the second using a function transform to compute the standard deviations.
    The optimizer is unaware that both objectives use the same set of
    realizations. To prevent redundant calculations, the evaluator should
    compute the results of the realizations once and return them for both
    objectives.

    Non-linear constraint values may potentially appear multiple times in the
    `constraint_results` matrix. For instance, to express the constraint $a <
    F_c \le b$, two constraints must be defined: $F_c \ge a$ and $F_c \le b$,
    sharing the same function value $F_c$ but differing in types and
    right-hand-sides (`ConstraintType.GE`/`ConstraintType.LE` and $a$/$b$). The
    run method should ensure that the calculated value of $F_c$ is the same in
    both cases, which is most efficiently achieved by evaluating $F_c$ only
    once.
"""

from ._concurrent import ConcurrentEvaluator, ConcurrentTask
from ._evaluator import Evaluator, EvaluatorContext, EvaluatorResult

__all__ = [
    "ConcurrentEvaluator",
    "ConcurrentTask",
    "Evaluator",
    "EvaluatorContext",
    "EvaluatorResult",
]



"""A class for running optimizations with script-based jobs."""

import json
import logging
import os
from collections import defaultdict
from pathlib import Path
from shutil import rmtree
from string import Template
from traceback import format_exception
from typing import (
    TYPE_CHECKING,
    Any,
    DefaultDict,
    Dict,
    List,
    Optional,
    TextIO,
    Union,
    no_type_check,
)

import numpy as np
from numpy.typing import NDArray
from parsl.app.app import bash_app
from parsl.providers.base import ExecutionProvider
from tabulate import tabulate

from ropt.config.plan import PlanConfig
from ropt.enums import EventType, OptimizerExitCode
from ropt.evaluator import EvaluatorContext
from ropt.evaluator.parsl import ParslEvaluator, State
from ropt.events import Event
from ropt.exceptions import ConfigError
from ropt.plan import OptimizerContext, Plan

from ._config import ScriptEvaluatorConfig, ScriptOptimizerConfig
from ._task import ScriptTask
from ._utils import _format_list, _get_function_files, _make_dict

if TYPE_CHECKING:
    from ropt.results import FunctionResults


@no_type_check
@bash_app()
def run_script(
    _: Any,  # noqa: ANN401
    script: str,
    stdout: TextIO,  # noqa: ARG001
    stderr: TextIO,  # noqa: ARG001
) -> str:
    return script


class ScriptOptimizer:
    """Optimizer class for running script-based optimization plans."""

    def __init__(
        self,
        config: Union[Dict[str, Any], ScriptOptimizerConfig],
        plan: Dict[str, Any],
        tasks: Dict[str, str],
        *,
        seed: Optional[int] = None,
    ) -> None:
        """Initialize the optimizer.

        The directory where the jobs run is constructed from by nesting
        directories according the `job_labels` tuple. These directories are
        located in the directory where the optimizer runs, which can be set
        using the `work_dir` parameter of the `run()` method.

        Args:
            config: Script optimizer configuration
            plan:   The optimization plan to run
            tasks:  A dictionary mapping task names to strings containing bash code
            seed:   Optional seed used by the optimization code
        """
        self._config = (
            config
            if isinstance(config, ScriptOptimizerConfig)
            else ScriptOptimizerConfig.model_validate(config)
        )
        self._plan_config = plan
        self._tasks = tasks
        self._seed = seed
        self._status: Dict[int, Any] = {}
        self._optimal_result: Optional[FunctionResults] = None

    def _set_logger(self) -> None:
        self._logger = logging.getLogger("ScriptBasedOptimizer")
        self._logger.setLevel(level=logging.INFO)
        handler = logging.FileHandler("optimizer.log")
        formatter = logging.Formatter("%(levelname)s: %(message)s")
        handler.setFormatter(formatter)
        self._logger.addHandler(handler)

    def _function(
        self,
        batch_id: int,
        job_idx: int,
        variables: NDArray[np.float64],
        context: EvaluatorContext,
    ) -> List[ScriptTask]:
        assert context.config.realizations.names is not None
        realization = context.config.realizations.names[context.realizations[job_idx]]
        if not isinstance(realization, int):
            msg = f"Realization name must be an integer: {realization}"
            raise ConfigError(msg)

        job_labels = tuple(
            label.format(batch=batch_id, realization=realization, job=job_idx)
            for label in self._config.job_labels
        )
        assert self._config.job_dir is not None
        path = self._config.job_dir / Path(*job_labels)
        objective_names, constraint_names = _get_function_files(context.config)
        objective_paths = tuple(path / name for name in objective_names)
        constraint_paths = tuple(path / name for name in constraint_names)

        var_path = path / "var-vector.npy"
        if all(
            path.exists() for path in (var_path, *objective_paths, *constraint_paths)
        ) and np.allclose(np.load(var_path), variables):
            msg = f"Batch {batch_id}, realization {realization}, job {job_idx}: SKIPPED"
            self._logger.info(msg)
            return [
                ScriptTask(
                    future=None,
                    objective_paths=objective_paths,
                    constraint_paths=constraint_paths,
                )
            ]

        if not self._tasks.items():
            return []

        rmtree(path, ignore_errors=True)
        Path.mkdir(path, parents=True)
        np.save(var_path, variables)

        assert context.config.variables.names is not None
        filename = path / self._config.var_filename.with_suffix(".json")
        var_dict = _make_dict(variables, context.config.variables.names)
        with filename.open("w", encoding="utf-8") as file_obj:
            json.dump(var_dict, file_obj, indent=2)

        tasks: List[ScriptTask] = []
        for task_name, script in self._tasks.items():
            substituted_script = Template(script).safe_substitute(
                work_dir=self._config.work_dir, realization=realization
            )
            tasks.append(
                ScriptTask(
                    name=task_name,
                    future=run_script(
                        tasks[-1].future if tasks else None,
                        f"cd {path}\n{substituted_script}",
                        stdout=((path / task_name).with_suffix(".stdout"), "a"),
                        stderr=((path / task_name).with_suffix(".stderr"), "a"),
                    ),
                    job_labels=job_labels,
                    realization=realization,
                ),
            )
        tasks[-1].objective_paths = objective_paths
        tasks[-1].constraint_paths = constraint_paths

        return tasks

    def _log_task(self, task: ScriptTask) -> None:
        if not task.logged:
            job_label = ", ".join(task.job_labels)

            if task.state == State.FAILED:
                assert task.exception is not None
                msg = f"{job_label}, {task.name}: FAILED\n"
                msg += "".join(
                    format_exception(
                        type(task.exception),
                        task.exception,
                        task.exception.__traceback__,
                    ),
                )
                self._logger.error(msg)
                task.logged = True

            if task.state == State.SUCCESS:
                msg = f"{job_label}, {task.name}: FINISHED"
                self._logger.info(msg)
                task.logged = True

    def _write_state_report(
        self, batch_id: int, jobs: Dict[int, List[ScriptTask]]
    ) -> None:
        states: DefaultDict[str, Dict[int, State]] = defaultdict(dict)
        for job_idx, tasks in jobs.items():
            for task in tasks:
                states[task.name][job_idx] = task.state
        table: List[Dict[str, Any]] = []
        for name, task_states in states.items():
            job_name = name
            for state in State:
                job_indices = [
                    idx for idx, item in task_states.items() if item == state
                ]
                if job_indices:
                    table.append(
                        {
                            "Task": job_name,
                            "State": state.value,
                            "Jobs": _format_list(job_indices),
                        },
                    )
                    job_name = ""
        with (self._config.work_dir / "status.txt").open(
            "w", encoding="utf-8"
        ) as file_obj:
            file_obj.write(f"Batch: {batch_id}\n\n")
            table_str = tabulate(table, headers="keys", tablefmt="simple")
            file_obj.write(f"{table_str}\n")

    def _update_current_state(
        self, batch_id: int, jobs: Dict[int, List[ScriptTask]]
    ) -> None:
        # Update the current batch
        states: DefaultDict[int, Dict[str, Any]] = defaultdict(dict)
        for job_idx, tasks in jobs.items():
            for task in tasks:
                states[job_idx][task.name] = {
                    "state": task.state.value,
                    "realization": task.realization,
                }
                self._log_task(task)
        self._status[batch_id] = states

    def _store_states(self) -> None:
        states_json = []
        for batch_id in sorted(self._status.keys()):
            states = self._status[batch_id]
            states_json.append(
                {
                    "batch_id": batch_id,
                    "jobs": [
                        {
                            "job": job_idx,
                            "tasks": [
                                {
                                    "name": task_name,
                                    "state": task_state["state"],
                                    "realization": task_state["realization"],
                                }
                                for task_name, task_state in states[job_idx].items()
                            ],
                        }
                        for job_idx in sorted(self._status[batch_id])
                    ],
                },
            )
        with (self._config.work_dir / "states.json").open(
            "w", encoding="utf-8"
        ) as file_obj:
            json.dump(states_json, file_obj, sort_keys=True, indent=4)

    def _monitor(self, batch_id: int, jobs: Dict[int, List[ScriptTask]]) -> None:
        self._write_state_report(batch_id, jobs)
        self._update_current_state(batch_id, jobs)
        self._store_states()

    def _log_exit_code(self, event: Event) -> None:
        msg = ""
        if event.exit_code == OptimizerExitCode.TOO_FEW_REALIZATIONS:
            msg = "Too few successful realizations: optimization stopped."
        elif event.exit_code == OptimizerExitCode.MAX_FUNCTIONS_REACHED:
            msg = "Maximum number of functions reached: optimization stopped."
        elif event.exit_code == OptimizerExitCode.USER_ABORT:
            msg = "Optimization plan aborted by the user."
        elif event.exit_code == OptimizerExitCode.OPTIMIZER_STEP_FINISHED:
            msg = "Optimization finished normally."
        if msg:
            if event.step_name is not None:
                msg = f"Step {event.step_name}: {msg}"
            self._logger.info(msg)

    def run(
        self,
        provider: Optional[ExecutionProvider] = None,
        evaluator_config: Optional[Union[Dict[str, Any], ScriptEvaluatorConfig]] = None,
    ) -> Plan:
        """Run the optimization."""
        cwd = Path.cwd()
        Path.mkdir(self._config.work_dir, parents=True, exist_ok=True)
        os.chdir(self._config.work_dir)

        if evaluator_config is None:
            evaluator_config = ScriptEvaluatorConfig()
        elif not isinstance(evaluator_config, ScriptEvaluatorConfig):
            evaluator_config = ScriptEvaluatorConfig.model_validate(evaluator_config)

        try:
            self._set_logger()
            with ParslEvaluator(
                self._function,
                polling=evaluator_config.polling,
                max_submit=evaluator_config.max_submit,
                max_threads=evaluator_config.max_threads,
            ).with_htex(
                provider=provider,
                htex_kwargs=evaluator_config.htex_kwargs,
                worker_restart=evaluator_config.worker_restart,
            ).with_monitor(
                self._monitor,
            ) as evaluator:
                context = OptimizerContext(evaluator=evaluator, seed=self._seed)
                config = PlanConfig.model_validate(self._plan_config)
                plan = Plan(config, context)
                plan.optimizer_context.events.add_observer(
                    EventType.FINISHED_OPTIMIZER_STEP, self._log_exit_code
                )
                plan.run()
        finally:
            os.chdir(cwd)

        return plan



"""This module implements the default optimizer step."""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Union

import numpy as np
from pydantic import BaseModel, ConfigDict

from ropt.config.enopt import EnOptConfig
from ropt.config.utils import Array2D  # noqa: TCH001
from ropt.ensemble_evaluator import EnsembleEvaluator
from ropt.enums import EventType, OptimizerExitCode
from ropt.exceptions import OptimizationAborted, PlanError
from ropt.plan import ContextUpdateResults
from ropt.plugins.plan.base import PlanStep
from ropt.results import FunctionResults

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ropt.config.plan import StepConfig
    from ropt.plan import Plan


class DefaultEvaluatorStepWith(BaseModel):
    """Parameters used by the default optimizer step.

    Optionally the initial variables to be used can be set from an context
    object.

    Attributes:
        config:   ID of the context object that contains the optimizer configuration
        update:   List of the objects that are notified of new results
        values:   Values to evaluate at
        metadata: Metadata to set in the results
    """

    config: str
    update: List[str] = []
    values: Optional[Union[str, Array2D]] = None
    metadata: Dict[str, str] = {}

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
        arbitrary_types_allowed=True,
    )


class DefaultEvaluatorStep(PlanStep):
    """The default evaluator step."""

    def __init__(self, config: StepConfig, plan: Plan) -> None:
        """Initialize a default evaluator step.

        Args:
            config: The configuration of the step
            plan:   The optimization plan that runs this step
        """
        super().__init__(config, plan)

        self._with = (
            DefaultEvaluatorStepWith.model_validate({"config": config.with_})
            if isinstance(config.with_, str)
            else DefaultEvaluatorStepWith.model_validate(config.with_)
        )

    def run(self) -> bool:
        """Run the evaluator step.

        Returns:
            Whether a user abort occurred.
        """
        config = self.plan.parse_value(self._with.config)
        if not isinstance(config, (dict, EnOptConfig)):
            msg = "No valid EnOpt configuration provided"
            raise PlanError(msg, step_name=self.step_config.name)
        self._enopt_config = EnOptConfig.model_validate(config)

        self.plan.optimizer_context.events.emit(
            EventType.START_EVALUATOR_STEP,
            config=self._enopt_config,
            step_name=self.step_config.name,
        )

        assert self.plan.optimizer_context.rng is not None
        ensemble_evaluator = EnsembleEvaluator(
            self._enopt_config,
            self.plan.optimizer_context.evaluator,
            self.plan.optimizer_context.result_id_iter,
            self.plan.optimizer_context.rng,
            self.plan.plugin_manager,
        )

        variables = self._get_variables()
        exit_code = OptimizerExitCode.EVALUATION_STEP_FINISHED

        try:
            results = ensemble_evaluator.calculate(
                variables, compute_functions=True, compute_gradients=False
            )
        except OptimizationAborted as exc:
            exit_code = exc.exit_code

        assert results
        assert isinstance(results[0], FunctionResults)
        if results[0].functions is None:
            exit_code = OptimizerExitCode.TOO_FEW_REALIZATIONS

        metadata = self.plan.metadata
        if metadata is None:
            metadata = {}
        if self.step_config.name is not None:
            metadata["step_name"] = self.step_config.name
        for key, expr in self._with.metadata.items():
            metadata[key] = self.plan.parse_value(expr)
        for item in results:
            item.metadata = metadata

        for obj_id in self._with.update:
            self.plan.update_context(
                obj_id,
                ContextUpdateResults(
                    step_name=self.step_config.name,
                    results=results,
                ),
            )

        self.plan.optimizer_context.events.emit(
            EventType.FINISHED_EVALUATOR_STEP,
            config=self._enopt_config,
            results=results,
            exit_code=exit_code,
            step_name=self.step_config.name,
        )

        return exit_code == OptimizerExitCode.USER_ABORT

    def _get_variables(self) -> NDArray[np.float64]:
        if self._with.values is not None:  # noqa: PD011
            parsed_variables = self.plan.parse_value(self._with.values)
            if isinstance(parsed_variables, FunctionResults):
                return parsed_variables.evaluations.variables
            if isinstance(parsed_variables, np.ndarray):
                return parsed_variables
            if parsed_variables is not None:
                msg = f"`{self._with.values} does not contain variables."  # noqa: PD011
                raise PlanError(msg, step_name=self.step_config.name)
        return self._enopt_config.variables.initial_values



"""Exceptions raised by `ropt`."""

from typing import Optional

from .enums import OptimizerExitCode


class ConfigError(Exception):
    """Raised when an configuration error occurs."""


class PlanError(Exception):
    """Raised when an error occurs in an optimization plan."""

    def __init__(
        self,
        message: str,
        *,
        step_name: Optional[str] = None,
        context_id: Optional[str] = None,
    ) -> None:
        """Initialize with a custom error message.

        Args:
            message:    Error message
            step_name:  Optional step name to add to the message
            context_id: Optional context ID to add to the message
        """
        if step_name is not None:
            msg = f"Step `{step_name}`: "
        elif context_id is not None:
            msg = f"Context object with ID `{context_id}`: "
        else:
            msg = ""
        super().__init__(msg + message)


class OptimizationAborted(Exception):  # noqa: N818
    """Raised when an optimization is aborted.

    When constructing the exception object an exit code must be passed that
    indicates the reason for aborting (see
    [`OptimizerExitCode`][ropt.enums.OptimizerExitCode]).
    """

    def __init__(self, exit_code: OptimizerExitCode) -> None:
        """Initialize an exception that aborts the optimization.

        Args:
            exit_code: The exit code indicating the reason for the abort
        """
        self.exit_code = exit_code
        super().__init__()



from typing import List, Optional

import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig
from ropt.plugins.function_transform.base import FunctionTransform


def _calculate_transformed_functions(  # noqa: PLR0913
    config: EnOptConfig,
    function_transforms: List[FunctionTransform],
    functions: NDArray[np.float64],
    realization_weights: Optional[NDArray[np.float64]],
    failed_realizations: NDArray[np.bool_],
    *,
    constraints: bool = False,
) -> NDArray[np.float64]:
    result: NDArray[np.float64] = np.empty(functions.shape[-1], dtype=np.float64)

    if constraints:
        assert config.nonlinear_constraints is not None
        transform_indices = config.nonlinear_constraints.function_transforms
    else:
        transform_indices = config.objective_functions.function_transforms

    if transform_indices is None:
        transform_indices = np.zeros(functions.shape[1], dtype=np.intc)

    for transform_idx, transform in enumerate(function_transforms):
        _add_transformed_functions(
            config,
            transform,
            functions,
            realization_weights,
            failed_realizations,
            transform_indices == transform_idx,
            result,
        )

    return result


def _add_transformed_functions(  # noqa: PLR0913
    config: EnOptConfig,
    transform: FunctionTransform,
    functions: NDArray[np.float64],
    realization_weights: Optional[NDArray[np.float64]],
    failed_realizations: NDArray[np.bool_],
    mask: NDArray[np.bool_],
    result: NDArray[np.float64],
) -> None:
    for idx in np.where(mask)[0]:
        weights = (
            config.realizations.weights
            if realization_weights is None
            else realization_weights[idx, ...]
        )
        weights = np.where(failed_realizations, 0.0, weights)
        weights /= weights.sum()
        result[idx] = transform.calculate_function(functions[..., idx], weights)


def _calculate_weighted_function(
    functions: NDArray[np.float64],
    weights: NDArray[np.float64],
    scales: NDArray[np.float64],
) -> NDArray[np.float64]:
    return np.array((weights * functions / scales).sum())
